# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Union, Tuple, List

import pandas as pd

from openstef_dbc.data_interface import _DataInterface


class Systems:
    def get_systems_near_location(
        self,
        location: Union[Tuple[float, float], str],
        radius: float = 15.0,
        quality: float = 0.75,
        freq: int = None,
        lag_systems: float = None,
    ) -> pd.DataFrame:
        """Retrieve all systems near a given location.
        Only return systems which exceed the quality, freq, lag_systems requirements.

        Parameters:
            - location: (lat, lon)
            - radius: float (km)
            - quality: float
            - freq: int (minutes)
            - lag_systems: float (minutes)

        Return:
            - pd.DataFrame(sid, lat, lon, region, distance)
        """

        # Define base query
        # use Great-circle calculation to determine systems within radius (6371 is radius of Earth)
        bind_params = {
            "lat": str(location[0]),
            "lon": str(location[1]),
            "quality": quality,
            "radius": str(radius),
        }
        query = """
                    SELECT `sid`, `lat`, `lon`,`region`, ( 6371 * acos( cos( radians(:lat) ) \
                * cos( radians( lat ) ) * cos( radians( lon ) - radians(:lon) ) + sin( radians(:lat) ) \
                * sin( radians( lat ) ) ) ) AS `distance` \
                FROM `systems`
                WHERE `qual` > ':quality'
                """

        # Extend query
        if freq is not None:
            bind_params["freq"] = str(freq)
            query += """ AND `freq` <= :freq"""
        if lag_systems is not None:
            query += """ AND `lagSystems` <= :quality"""

        # Limit radius to given input radius
        query += """ HAVING `distance` < :radius ORDER BY `distance`;"""

        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)
        return result

    def get_systems_by_pid(
        self, pid: int, return_list: bool = False
    ) -> Union[pd.DataFrame, List[dict]]:
        """Get all prediction job systems for a given prediction job id.

            The `systems` table will be joint with the `predicions_systems` table
            in order to return all columns of `systems` and `predictions_systems`.
            The latter will add the `factor` column. The factor determines wether or
            not a systems should be added or subtracted for this pid.

        Args:
            pid (Union[int, str): Prediction job id
            return_list (Optional[bool]): Return a List[dict] of prediction job systems.
                Defaults to False.
        Returns:
            Union[pandas.DataFrame, List[dict]]: Prediction job systems
        """

        query = """
        SELECT * from systems
        INNER JOIN predictions_systems
            ON predictions_systems.system_id=systems.sid
        WHERE predictions_systems.prediction_id=:pid
        """

        systems = _DataInterface.get_instance().exec_sql_query(
            query, params={"pid": pid}
        )

        if return_list is False:
            return systems

        return systems.to_dict("records")

    def get_pv_systems_with_incorrect_location(self) -> pd.DataFrame:
        """Get PV systems with missing (or incorrect) lat/lon"""
        query = """
            SELECT * FROM `systems`
            WHERE `sid` LIKE 'pv_%' AND (`lat` IS NULL OR `lon` IS NULL)
        """
        return _DataInterface.get_instance().exec_sql_query(query)

    def get_random_pv_systems(
        self, autoupdate: int = 1, limit: int = None
    ) -> pd.DataFrame:
        limit_query = ""

        bind_params = {"autoupdate": autoupdate}

        if limit is not None:
            bind_params["limit"] = limit
            limit_query = f"LIMIT :limit"

        query = f"""
            SELECT sid, qual, freq, lag
            FROM systems
            WHERE left(sid, 3) = 'pv_' AND autoupdate = :autoupdate
            ORDER BY RAND() {limit_query}
        """

        return _DataInterface.get_instance().exec_sql_query(query, bind_params)

    def get_api_key_for_system(self, sid: str) -> str:
        """Get (sysetm) API key that is connected to a given system.

        Args:
            sid (str): The system ID

        Returns:
            str: The API key that is connected to the given system.
        """
        bind_params = {"system": sid}

        query = """
            SELECT 
                sa.apiKey 
            FROM `systems` as s 
            LEFT JOIN `systemsApiKeys` as sa ON s.measurements_customer_api_key_id = sa.id 
            WHERE s.sid = :system;
        """

        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)
        if isinstance(result, pd.DataFrame) and result.empty:
            return ""
        else:
            return result["apiKey"][0]

    def get_api_keys_for_systems(self, sids: list[str]) -> list[str]:
        """Get (sysetm) API keys that are connected to a given list of systems.

        Args:
            sid (str): The list with system IDs

        Returns:
            list(str): The API keys that are connected to the given systems in the list.
        """
        api_keys = []
        for sid in sids:
            api_keys.append(self.get_api_key_for_system(sid))
        return list(set(api_keys))



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0


def add_application_metadata(logger, method_name, event_dict):
    """Add additional API metadata to the passed event dictionary.
    Note:
        The unused arguments are still passed by the caller and need to be caught.
    Args:
        logger: wrapped logger object
        method_name: name of the wrapper method
        event_dict: current context and event
    Returns:
        Dictionary with possibly extra fields
    """
    #    if "app_version" not in event_dict:
    #        event_dict["app_version"] = get_setting("APP_VERSION")

    return event_dict


def rename_forbidden_keys(logger, method_name, event_dict):
    """Certain keys should not be used for logging, as they are already
    claimed by kibana. Not respecting the Kibana-claimed-keys results in
    ommitted logging at cluster level.
    This function adds a prefix 'log.' to the forbidden keys


    Example:
        event_dict = dict(event='textmessage', specific.key='toedeloe') ->
        event_dict = dict(log.event='textmessage', specific.key='toedeloe')
    """
    forbidden_keys = ["event"]

    # For all keys in forbidden keys AND event_dict, add prefix 'log.'
    for key in set(forbidden_keys).intersection(set(event_dict.keys())):
        event_dict[f"log.{key}"] = event_dict.pop(key)

    return event_dict



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
import pytz
from typing import Union, List
from datetime import datetime, timedelta

import pandas as pd
import re

from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.utils import floor_to_closest_time_resolution, parse_influx_result

FIELDS_OF_INTEREST = [
    "forecast",
    "stdev",
    "forecast_other",
    "forecast_solar",
    "forecast_wind_on_shore",
]


class Predictions:
    def get_predicted_load(
        self,
        pj: dict,
        start_time: datetime = None,
        end_time: datetime = None,
    ) -> pd.Series:
        """Get historic load predictions for given pid.

        This functions ignores component predictions that result from energy splitting.

        Args:
            pj (dict): Prediction job
            start_time (datetime): Start time to retrieve the historic load prediction.
            end_time (datetime): End time to retrieve the historic load prediction.

        Returns:
            pandas.Series: Forecast column with the predicted values

        """
        # Apply default parameters if none are provided
        if start_time is None:
            start_time = datetime.utcnow()
        if end_time is None:
            end_time = datetime.utcnow() + timedelta(days=2)
        else:
            end_time = end_time + timedelta(
                minutes=15
            )  # To make the ouput the same as the original influx client.

        # Round end_time to last resolution_minutes interval
        end_time = floor_to_closest_time_resolution(end_time, pj["resolution_minutes"])

        bind_params = {
            "pid": str(pj["id"]),
            "_start": start_time.astimezone(pytz.UTC),
            "_stop": end_time.astimezone(pytz.UTC),
        }

        query = f"""
            from(bucket: "forecast_latest/autogen")
                |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                |> filter(fn: (r) => 
                    r._measurement == "prediction")
                |> filter(fn: (r) => 
                    r._field == "forecast" or r._field == "stdev") 
                |> filter(fn: (r) => 
                    r.type != "est_demand" and  r.type != "est_wind" and  r.type != "est_pv") 
                |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
        """
        # Query the database
        result = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        # Return result
        if not result.empty:
            # Shifting is needed to output the same as with the old influx client
            return parse_influx_result(result).shift(-15, freq="min")
        else:
            return pd.Series()

    def get_forecast_quality(
        self, pj: dict, start_time: datetime = None, end_time: datetime = None
    ) -> pd.Series:
        """Get forecast quality for historic load predictions for given pid.

        Args:
            pj (dict): Prediction job
            start_time (datetime): Start time to retrieve the historic load prediction.
            end_time (datetime): End time to retrieve the historic load prediction.

        Returns:
            pandas.Series: Quality column with the quality of the predicted values

        """
        if start_time is None:
            start_time = datetime.utcnow()
        if end_time is None:
            end_time = datetime.utcnow() + timedelta(days=2)

        bind_params = {
            "pid": str(pj["id"]),
            "_start": start_time.astimezone(pytz.UTC),
            "_stop": end_time.astimezone(pytz.UTC),
        }

        query = f"""
        from(bucket: "forecast_latest/autogen")
            |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
            |> filter(fn: (r) => 
                r._measurement == "prediction")
            |> filter(fn: (r) => 
                r._field == "quality" )                 
            |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
        """

        # Query the database
        result = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        # Return result
        if not result.empty:
            return parse_influx_result(result, aggfunc="first")
        else:
            return pd.Series()

    def get_prediction_including_components(
        self,
        pj: dict,
        start_time: datetime = None,
        end_time: datetime = None,
        quantiles: bool = False,
    ):
        """Get historic load predictions for given pid including component forecasts.

        Args:
            pj (dict): Prediction job
            start_time (datetime): Start time  to retrieve the historic load prediction.
            end_time (datetime): End timeto retrieve the historic load prediction.
            quantiles: (bool): Indicates wheter quantiles should be retrieved as well.

        Returns:
            pandas.DataFrame: Dataframe with the total forecast,
            the components forecasts and the quantiles if they are requested.

        """
        # Apply default parameters if none are provided
        if start_time is None:
            start_time = datetime.utcnow()
        if end_time is None:
            end_time = datetime.utcnow() + timedelta(days=2)

        bind_params = {
            "pid": str(pj["id"]),
            "_start": start_time.astimezone(pytz.UTC),
            "_stop": end_time.astimezone(pytz.UTC),
        }
        quantile_section = (
            """or r._field == "quantile_P"""
            + """" or r._field == "quantile_P""".join(
                [f"{quantile * 100:02.0f}" for quantile in pj["quantiles"]]
            )
            + '"'
        )
        fields_of_interest_section = (
            'r._field == "' + '" or r._field == "'.join(FIELDS_OF_INTEREST) + '"'
        )

        if quantiles:
            query = f"""
            from(bucket: "forecast_latest/autogen")
                |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                |> filter(fn: (r) => 
                    r._measurement == "prediction")
                |> filter(fn: (r) => 
                   {fields_of_interest_section} {quantile_section})                 
                |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
        """

        else:
            query = f"""
                from(bucket: "forecast_latest/autogen")
                    |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                    |> filter(fn: (r) => 
                        r._measurement == "prediction")
                    |> filter(fn: (r) => 
                        {fields_of_interest_section})                 
                    |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                    |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
            """

        # Query the database
        result = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        # Return result
        if not result.empty:
            # Shifting is needed to output the same as with the old influx client
            return parse_influx_result(result).shift(-15, freq="min")
        else:
            return pd.Series()

    def get_predicted_load_tahead(
        self,
        pj: dict,
        start_time: datetime = None,
        end_time: datetime = None,
        t_ahead: Union[List[str], str] = None,
        component: bool = False,
    ) -> pd.DataFrame:
        """Get historic load predictions for given pid and t_ahead.

        This functions ignores component predictions that result from energy splitting.

        Args:
            pj (dict): Prediction job
            start_time (datetime): Start time  to retrieve the historic load prediction.
            end_time (datetime): End timeto retrieve the historic load prediction.
            t_ahead (String, list of strings or None):
                Specifies specific t_ahead(s) for which to retrieve historic predictions,
                if none all available t_aheads will be returned.

        Returns:
            pandas.DataFrame: object with a column for each requested t_ahead

        """
        # Apply default parameters if none are provided
        if start_time is None:
            start_time = datetime.utcnow() - timedelta(days=1)
        if end_time is None:
            end_time = datetime.utcnow()

        # If no t_ahead are provided ask influx for all t_ahead available
        if t_ahead is None:
            if component:
                bind_params = {
                    "pid": str(pj["id"]),
                    "_start": start_time.astimezone(pytz.UTC),
                    "_stop": end_time.astimezone(pytz.UTC),
                }
                query = f"""
                    from(bucket: "forecast_latest/autogen")
                        |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                        |> filter(fn: (r) => 
                            r._measurement == "prediction_tAheads")
                        |> filter(fn: (r) => 
                            r._field == "forecast_solar" or r._field == "stdev") 
                        |> filter(fn: (r) => 
                            r.type != "est_demand" and  r.type != "est_wind" and  r.type != "est_pv") 
                        |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                        |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
                """
            else:
                bind_params = {
                    "pid": str(pj["id"]),
                    "_start": start_time,
                    "_stop": end_time,
                }
                query = f"""
                    from(bucket: "forecast_latest/autogen")
                        |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                        |> filter(fn: (r) => 
                            r._measurement == "prediction_tAheads")
                        |> filter(fn: (r) => 
                            r._field == "forecast" or r._field == "stdev") 
                        |> filter(fn: (r) => 
                            r.type != "est_demand" and  r.type != "est_wind" and  r.type != "est_pv") 
                        |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                        |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
                """

        # For a selection of t_aheads a custom query is generated
        else:
            # If string convert to single item list and continue
            if isinstance(t_ahead, str):
                t_ahead = [t_ahead]

            # Raise exception when input is not a list of strings or a string.
            elif not all(isinstance(item, str) for item in t_ahead) and not isinstance(
                t_ahead, str
            ):
                raise ValueError("Could not interpret t_ahead argument!")

            # Placeholder string
            t_aheads = ""
            # Loop over requested t_aheads and add them to the query
            for item in t_ahead:
                # Get rid of hour symbols
                item = re.sub("[hH]", "", item)
                # Make string for this t_ahead
                t_aheads = t_aheads + """ "tAhead" = '{}' OR""".format(str(item) + ".0")

            # Get rid of last OR
            t_aheads = t_aheads[0:-2]

            # Make query for a selection of t_aheads
            bind_params = {
                "pid": str(pj["id"]),
                "_start": start_time.astimezone(pytz.UTC),
                "_stop": end_time.astimezone(pytz.UTC),
            }
            t_aheads_section = '" or r.taheads == "'.join(t_aheads)

            query = f"""
                from(bucket: "forecast_latest/autogen")
                    |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                    |> filter(fn: (r) => 
                        r._measurement == "prediction_tAheads")
                    |> filter(fn: (r) => 
                        r._field == "forecast" or r._field == "stdev") 
                    |> filter(fn: (r) => 
                        r.type != "est_demand" and  r.type != "est_wind" and  r.type != "est_pv") 
                    |> filter(fn: (r) => r.pid == "{bind_params["pid"]}")
                    |> filter(fn: (r) => r.taheads == "{t_aheads_section}")
                    |> aggregateWindow(every: {pj["resolution_minutes"]}m, fn: mean)
            """

        # Query the database
        result = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        # Return result
        if not result.empty:
            # Shifting is needed to output the same as with the old influx client
            result = parse_influx_result(result, aditional_indices=["tAhead"]).shift(
                -15, freq="min"
            )
        else:
            return pd.Series()

        # Convert to pandas DataFrame with a column for each tAhead
        predicted_load = result.pivot(columns="tAhead")

        predicted_load.columns = [
            "_".join(col) + "h" for col in predicted_load.columns.values
        ]

        # Return result
        return predicted_load



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Union, Tuple
from datetime import datetime, timedelta

import pandas as pd
import numpy as np

from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.weather import Weather
from openstef_dbc.services.systems import Systems
from openstef_dbc.services.ems import Ems
from openstef_dbc.services.predictor import Predictor


# TODO check if this functionality should be transfered to a data preparation step for energy splitting in ktprognoses
class Splitting:
    def get_energy_split_coefs(self, pj: dict, mean: bool = False) -> dict:
        """
            Method to retrieve latest energy splitting coefficients from the database.
            If mean is passed as True the average values are given for the last 180 days.
        Args:
            pj: (dict) Prediction job
            mean: (bool), if true the average of splitting coefficients in the last 180 days are returned.

        Returns:
            (dict) with latest splitting coefficients, or average vales of the last 180 days are returned

        """

        # Retrieve the average values of the coefficients of the last 180 if requested
        if mean:
            start_date = datetime.utcnow() - timedelta(days=180)
            bind_params = {"pid": pj["id"], "dstart": start_date.isoformat()}
            query = (
                "SELECT ec.coef_name, AVG(ec.coef_value) FROM energy_split_coefs as ec "
                "WHERE ec.pid = :pid AND ec.created > :dstart GROUP BY ec.coef_name "
            )
        # Retrieve latest coefficients otherwise
        else:
            bind_params = {"pid": pj["id"]}
            query = (
                "SELECT ec.coef_name,ec.coef_value FROM energy_split_coefs as ec WHERE  ec.pid = :pid "
                "AND ec.created = (SELECT max(energy_split_coefs.created) from energy_split_coefs "
                "WHERE energy_split_coefs.pid = :pid)"
            )
        # Execute query
        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)

        # Make output dict
        if not result.empty:
            result = result.set_index("coef_name")
            if mean:
                result = result.to_dict()["AVG(ec.coef_value)"]
            else:
                result = result.to_dict()["coef_value"]
        else:
            result = {}

        return result

    def get_wind_ref(
        self,
        location: Union[Tuple[float, float], str],
        datetime_start: datetime,
        datetime_end: datetime,
        country: str = "NL",
        forecast_resolution: str = "15min",
        source: str = "optimum",
    ) -> pd.DataFrame:
        """Function that gets windspeed data from the influx database and converts it to windref data suitable
        for splitting energy.
        Parameters:
            location: str, country or location
            datetime_start: datetime, start time of required windref
            datetime_end: datetime, end time of required windref
            country: str, default is "NL"
        Output:
            windref: pandas dataframe containing the windref data"""

        # Get weather information from the influx database
        wind_speed = Weather().get_weather_data(
            location,
            ["windspeed_100m"],
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            source=source,
            resolution=forecast_resolution,
            country=country,
        )

        wind_ref = self._calculate_windspeed_at_hubheight(
            self, windspeed=wind_speed["windspeed_100m"], fromheight=100, hub_height=100
        )
        wind_ref = wind_ref / np.abs(np.amax(wind_ref))

        result = pd.DataFrame()
        result["windspeed"] = wind_ref

        # Return result
        return result

    def _get_solar_ref(
        self,
        location: Union[Tuple[float, float], str],
        datetime_start: datetime = None,
        datetime_end: datetime = None,
        forecast_resolution: str = "15min",
    ) -> pd.DataFrame:
        """Function that gets PV data from the influx database and converts it to solar_ref data suitable
        for splitting energy.
        Parameters:
            location: list, list of lat and lon of desired location
            datetime_start: datetime, start time of required solar_ref
            datetime_end: datetime, end time of required solar_ref
        Output:
            solar_ref: pandas dataframe containing the solar_ref data"""

        if datetime_start is None:
            datetime_start = datetime.utcnow().date() - timedelta(14)
        if datetime_end is None:
            datetime_end = datetime.utcnow().date() + timedelta(3)

        # Look for PV systems near desired location
        systems = Systems().get_systems_near_location(location, quality=0.95, freq=5)

        if len(systems) == 0:
            systems = Systems().get_systems_near_location(
                location, radius=30, quality=0.95, freq=5
            )

        # Get PV data from the archive table
        pvdata = Ems().get_load_sid(
            sid=systems["sid"],
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            forecast_resolution=forecast_resolution,
            aggregated=True,
            average_output=True,
        )
        pvdata.rename(columns={"load": "aggregated"}, inplace=True)

        # Normalize
        pvdata.aggregated = pvdata.aggregated / np.percentile(pvdata.aggregated, 99.0)

        # Return results
        return pvdata[["aggregated"]]

    def get_input_energy_splitting(
        self,
        pj: dict,
        datetime_start: datetime = None,
        datetime_end: datetime = None,
        forecast_resolution: str = "15min",
        ignore_factor: bool = False,
        country: str = "NL",
        source: str = "optimum",
    ) -> pd.DataFrame:
        if datetime_start is None:
            datetime_start = datetime.utcnow() - timedelta(days=90)
        if datetime_end is None:
            datetime_end = datetime.utcnow()

        # Get standard load profiles (StandaardJaarVerbruik in Dutch)
        sjv = Predictor().get_load_profiles(
            datetime_start, datetime_end, forecast_resolution
        )

        # Get windpower reference
        wind_ref = self.get_wind_ref(
            (pj["lat"], pj["lon"]),
            datetime_start,
            datetime_end,
            country,
            forecast_resolution,
            source,
        )

        # Get load data
        load = Ems().get_load_pid(
            pj["id"],
            datetime_start,
            datetime_end,
            forecast_resolution,
            ignore_factor=ignore_factor,
        )

        # Get solar (PV) power reference
        solar_ref = self._get_solar_ref(
            location=(pj["lat"], pj["lon"]),
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            forecast_resolution=forecast_resolution,
        )

        # Resample to forecast_resolution
        solar_ref = solar_ref.resample(forecast_resolution).mean()

        # Invert solar_ref and windref as electricity is produced and not consumed
        solar_ref *= -1
        wind_ref *= -1

        # Merge solar_ref and windref
        input_split_function = load.merge(
            wind_ref, left_index=True, right_index=True, how="outer"
        )
        input_split_function = input_split_function.merge(
            solar_ref, left_index=True, right_index=True, how="outer"
        )
        input_split_function = input_split_function.merge(
            sjv, left_index=True, right_index=True, how="outer"
        )

        # Drop rows with duplicate indices
        input_split_function = input_split_function[
            ~input_split_function.index.duplicated()
        ]

        # Rename columns to match their contents and the requirements of FindComponents()
        input_split_function.rename(columns={"windspeed": "wind_ref"}, inplace=True)
        input_split_function.rename(columns={"aggregated": "pv_ref"}, inplace=True)

        # Replace infs with nans and drop all rows with Nans
        input_split_function.replace([np.inf, -np.inf], np.nan)
        input_split_function.dropna(inplace=True)

        return input_split_function

    @staticmethod
    def _calculate_windspeed_at_hubheight(
        self,
        windspeed: Union[float, pd.Series],
        fromheight: float = 10.0,
        hub_height: float = 100.0,
    ) -> pd.Series:
        """
        function that extrapolates a wind from a certain height to 100m
        According to the wind power law (https://en.wikipedia.org/wiki/Wind_profile_power_law)

        input:
            - windspeed: float OR pandas series of windspeed at height = height
            - fromheight: height (m) of the windspeed data. Default is 10m
            - hubheight: height (m) of the turbine
        returns:
            - the windspeed at hubheight."""
        alpha = 0.143

        if not isinstance(windspeed, (np.ndarray, float, int, pd.Series)):
            raise TypeError(
                "The windspeed is not of the expected type!\n\
                            Got {}, expected np.ndarray, pd series or numeric".format(
                    type(windspeed)
                )
            )

        try:
            if any(windspeed < 0):
                raise ValueError(
                    "The windspeed cannot be negative, as it is the lenght of a vector"
                )
        except TypeError:
            if windspeed < 0:
                raise ValueError(
                    "The windspeed cannot be negative, as it is the lenght of a vector"
                )
            windspeed = abs(windspeed)

        return windspeed * (hub_height / fromheight) ** alpha



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

import json
from typing import List, Optional, Union
from pydantic import ValidationError
import pandas as pd

from openstef.data_classes.prediction_job import PredictionJobDataClass

from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.log import logging
from openstef_dbc.services.systems import Systems


class PredictionJobRetriever:
    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)

    def get_prediction_job(
        self, pid: int, model_type: str = None, is_active: int = None
    ) -> PredictionJobDataClass:
        """Get prediction job for a given pid from the database.

        Args:
            pid (int): Id of prediction job
            model_type (str, optional): Model type. Defaults to None.
            is_active (int, optional): Active job. Defaults to None.

        Returns:
            PredictionJobDataClass: Prediction job class
        """

        query = self.build_get_prediction_jobs_query(
            pid=pid, model_type=model_type, is_active=is_active
        )
        result = _DataInterface.get_instance().exec_sql_query(query)

        if result.size == 0:
            raise ValueError(f"No prediction job found with id '{pid}'")

        # Convert to dictionary
        prediction_job_dict = result.to_dict("records")[0]

        # Add description
        prediction_job_dict = self._add_description_to_prediction_job(
            prediction_job_dict
        )

        # Add quantiles
        prediction_job_dict = self._add_quantiles_to_prediction_job(prediction_job_dict)

        prediction_job = self._create_prediction_job_object(prediction_job_dict)
        return prediction_job

    def get_prediction_jobs(
        self,
        model_type: Optional[str] = None,
        is_active: int = 1,
        only_ato: bool = False,
        limit: Optional[int] = None,
    ) -> List[PredictionJobDataClass]:
        """Get all prediction jobs from the database.

        Args:
            model_type (str): Only retrieve jobs with specified modeltype (e.g. 'xgb').
                if None, all jobs are retrieved
            is_active (int): Only retrieve jobs where active == is_active.
                if None, all jobs are retrieved
            only_ato (bool): Only retrieve ATO jobs
            limit (int): Limit the number of jobs to given value.

        Returns:
            List[PredictionJobDataClass]: List of prediction jobs
        """
        query = self.build_get_prediction_jobs_query(
            model_type=model_type,
            is_active=is_active,
            only_ato=only_ato,
            limit=limit,
        )
        prediction_jobs = self._get_prediction_jobs_query_results(query)
        prediction_jobs = self._add_quantiles_to_prediction_jobs(prediction_jobs)
        return prediction_jobs

    def get_prediction_jobs_wind(self):
        query = """
            SELECT
                p.id, 
                p.forecast_type, 
                p.model, 
                p.horizon_minutes, 
                p.resolution_minutes,
                p.name,
                min(s.sid) as sid,
                w.lat as lat,
                w.lon as lon,
                w.turbine_type,
                w.n_turbines,
                w.hub_height
            FROM predictions as p
            LEFT JOIN predictions_systems as ps ON p.id = ps.prediction_id
            LEFT JOIN systems as s ON s.sid = ps.system_id
            LEFT JOIN windspecs as w ON p.id = w.pid
            WHERE p.forecast_type = 'wind' AND p.active = 1
            GROUP BY p.id
        """

        # Retrieve prediction jobs from database
        prediction_jobs = self._get_prediction_jobs_query_results(query)

        return prediction_jobs

    def get_prediction_jobs_solar(self):
        query = """
            SELECT
                p.id, 
                p.forecast_type, 
                p.model, 
                p.horizon_minutes, 
                p.resolution_minutes,
                p.name,
                min(s.lat) as lat,
                min(s.lon) as lon,
                min(s.sid) as sid,
                ss.lat,
                ss.lon,
                ss.radius,
                ss.peak_power
            FROM predictions as p
            LEFT JOIN predictions_systems as ps ON p.id = ps.prediction_id
            LEFT JOIN systems as s ON s.sid = ps.system_id
            LEFT JOIN solarspecs as ss ON p.id = ss.pid
            WHERE p.forecast_type = 'solar' AND p.active = 1
            GROUP BY p.id
        """

        # Retrieve prediction jobs from database
        prediction_jobs = self._get_prediction_jobs_query_results(query)

        return prediction_jobs

    def _get_prediction_jobs_query_results(
        self, query: str
    ) -> List[PredictionJobDataClass]:
        """Get prediction jobs using a query to the database."""
        results = _DataInterface.get_instance().exec_sql_query(query)
        if len(results) == 0:
            return []

        # Convert to list of dictionaries
        prediction_jobs = results.to_dict("records")

        # Add description to all prediction jobs
        prediction_jobs = self._add_description_to_prediction_jobs(prediction_jobs)
        prediction_jobs = [
            self._create_prediction_job_object(prediction_job)
            for prediction_job in prediction_jobs
        ]

        return prediction_jobs

    def _create_prediction_job_object(self, pj: dict) -> PredictionJobDataClass:
        """Create an object for the prediction job from a dictionary

        Args:
            pj (dict): dictionary with the attributes from the prediction job

        Returns:
            prediction_job_object (object): data class of the prediction job
        """
        try:
            # Change the typ column to forecast_type
            if "typ" in pj:
                pj["forecast_type"] = pj.pop("typ")
            if pj["train_horizons_minutes"]:
                pj["train_horizons_minutes"] = json.loads(pj["train_horizons_minutes"])
            prediction_job_object = PredictionJobDataClass(**pj)
        except ValidationError as e:
            errors = e.errors()
            self.logger.error(
                "Error occurred while converting to data class",
                pid=pj["id"],
                error=errors,
            )
            raise AttributeError(e)
        return prediction_job_object

    def _add_description_to_prediction_job(
        self, prediction_job: PredictionJobDataClass
    ) -> PredictionJobDataClass:
        return self._add_description_to_prediction_jobs([prediction_job])[0]

    def _add_quantiles_to_prediction_job(
        self, prediction_job: PredictionJobDataClass
    ) -> PredictionJobDataClass:
        return self._add_quantiles_to_prediction_jobs([prediction_job])[0]

    def get_pids_for_api_key(self, api_key: str) -> list[int]:
        """Get all pids that belong to a given API key.

        Args:
            api_key (str): The API key

        Returns:
            list[int]: dList of pids
        """
        bind_params = {"apiKey": api_key}
        query = """
            SELECT
                p.id 
            FROM `customersApiKeys` as cak 
            LEFT JOIN `customers` as cu ON cak.cid = cu.id 
            LEFT JOIN `customers_predictions` as cp ON cu.id = cp.customer_id 
            LEFT JOIN `predictions` as p ON p.id = cp.prediction_id 
            WHERE cak.apiKey = :apiKey 
        """
        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)
        if isinstance(result, pd.DataFrame) and result.empty:
            return []
        else:
            return result["id"].to_list()

    def get_pids_for_api_keys(self, api_keys: list[str]) -> list[int]:
        """Get all pids that belong to a given list of API keys.

        Args:
            api_key (str): The list with API keys

        Returns:
            list[int]: List of pids
        """
        pids = []
        for api_key in api_keys:
            pids.append(self.get_pids_for_api_key(api_key))
        return list(set(pids))

    def get_eans_for_pids(self, pids: list[int]) -> dict[int : list[str]]:
        """Get EANs that are connected to a prediction.

        Args:
            pid (int): The prediction ID

        Returns:
            dict{int:list[str]}: The pid to eans mapping
        """
        eans = {}
        for pid in pids:
            eans.update({pid: eans.append(self.get_ean_for_pid(pid))})
        return eans

    def get_ean_for_pid(self, pid: int) -> list[str]:
        """Get EAN that is connected to a prediction

        Args:
            pid (int): The prediction ID

        Returns:
            str: The corresponding EAN code
        """
        bind_params = {"pid": pid}
        query = """
            SELECT
                p.ean 
            FROM `predictions` as p  
            WHERE p.id = :pid 
        """
        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)
        if isinstance(result, pd.DataFrame) and result.empty:
            return []
        else:
            return result["ean"].to_list()

    @classmethod
    def _add_quantiles_to_prediction_jobs(
        cls, prediction_jobs: List[PredictionJobDataClass]
    ) -> List[PredictionJobDataClass]:
        prediction_job_ids = [pj["id"] for pj in prediction_jobs]
        prediction_jobs_ids_str = ", ".join([f"'{p}'" for p in prediction_job_ids])

        query = f"""
            SELECT p.id AS prediction_id, qs.quantiles
            FROM quantile_sets AS qs
            JOIN (
                predictions_quantile_sets AS pq,
                predictions as p
            )
            WHERE
                p.id = pq.prediction_id AND
                qs.id = pq.quantile_set_id AND
                p.id IN  ({prediction_jobs_ids_str})
            ORDER BY prediction_id
            ;
        """
        result = _DataInterface.get_instance().exec_sql_query(query)

        prediction_job_quantiles = {}

        # get quantiles for every prediction job
        for _, row in result.iterrows():
            pid = row["prediction_id"]

            if pid not in prediction_job_quantiles:
                prediction_job_quantiles[pid] = []

            prediction_job_quantiles[pid] += json.loads(row["quantiles"])

        # add quantiles to prediction job
        for prediction_job in prediction_jobs:
            pid = prediction_job["id"]
            # add quantiles if any
            if pid in prediction_job_quantiles:
                prediction_job["quantiles"] = sorted(prediction_job_quantiles[pid])
                continue
            # add empty list if none (this should not actually happen)
            prediction_job["quantiles"] = []

        return prediction_jobs

    @classmethod
    def build_get_prediction_jobs_query(
        cls,
        pid: Union[int, str, List[int], List[str], None] = None,
        model_type: Union[str, List[str], None] = None,
        is_active: Optional[int] = None,
        only_ato: bool = False,
        limit: Optional[int] = None,
    ) -> str:
        where_condition = []

        if pid is not None:
            where_condition.append(
                PredictionJobRetriever._build_pid_where_condition(pid)
            )

        if model_type is not None:
            where_condition.append(
                PredictionJobRetriever._build_model_type_where_condition(model_type)
            )

        if is_active is not None:
            where_condition.append(
                PredictionJobRetriever._build_active_where_condition(is_active)
            )

        if only_ato:
            where_condition.append("`name` LIKE 'ATO%' AND `name` NOT LIKE '%HS%'")

        where_clause = ""
        limit_clause = ""

        if len(where_condition) > 0:
            where_clause = f"WHERE {' AND '.join(where_condition)}"

        if limit:
            limit_clause = f"LIMIT {limit}"

        query = f"""
            SELECT
                p.id, 
                p.ean,
                p.name,
                p.forecast_type, 
                p.model, 
                p.horizon_minutes,
                p.train_horizons_minutes,
                p.resolution_minutes,
                p.train_components,
                min(s.lat) as lat,
                min(s.lon) as lon
            FROM predictions as p
            LEFT JOIN
                predictions_systems as ps ON p.id = ps.prediction_id
            LEFT JOIN
                systems as s ON s.sid = ps.system_id
            {where_clause}
            GROUP BY p.id
            {limit_clause};
        """
        return query

    @staticmethod
    def _add_description_to_prediction_jobs(
        prediction_jobs: List[PredictionJobDataClass],
    ) -> List[PredictionJobDataClass]:
        for prediction_job in prediction_jobs:
            systems = Systems().get_systems_by_pid(
                pid=prediction_job["id"], return_list=True
            )
            systems_str = "+".join([s["system_id"] for s in systems])
            prediction_job["description"] = systems_str

        return prediction_jobs

    @staticmethod
    def _build_pid_where_condition(pid: int) -> str:
        if isinstance(pid, list):
            in_values = ", ".join([f"'{p}'" for p in pid])
        elif isinstance(pid, int) or isinstance(pid, str):
            in_values = f"'{pid}'"
        else:
            raise ValueError("pid should be int, str, list of int or list of str")
        return f"p.id IN ({in_values})"

    @staticmethod
    def _build_model_type_where_condition(model_type: Union[list, str]) -> str:
        if isinstance(model_type, list):
            in_values = ", ".join([f"'{v}'" for v in model_type])
        elif isinstance(model_type, str):
            in_values = f"'{model_type}'"
        else:
            raise ValueError("model_type should be str or list of str")
        return f"p.model IN ({in_values})"

    @staticmethod
    def _build_active_where_condition(active: int) -> str:
        # make sure we always get 0 or 1 (anything not 0 -> 1)
        active = int(active != 0)
        return f"p.active = {active}"



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

import geopy
from influxdb_client import InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS
import pandas as pd
import numpy as np
import requests
import sqlalchemy
from sqlalchemy import text, bindparam
from openstef_dbc import Singleton
from openstef_dbc.ktp_api import KtpApi
from openstef_dbc.log import logging
from enum import Enum

# Define abstract interface


class _DataInterface(metaclass=Singleton):
    def __init__(self, config):
        """Generic data interface.

        All connections and queries to the InfluxDB database, SQL databases and
        influx API are governed by this class.

        Args:
            config: Configuration object. with the following attributes:
                api_username (str): API username.
                api_password (str): API password.
                api_admin_username (str): API admin username.
                api_admin_password (str): API admin password.
                api_url (str): API url.
                influxdb_token (str): Token to authenticate to InfluxDB.
                influxdb_host (str): InfluxDB host.
                influxdb_port (int): InfluxDB port.
                influx_organization (str): InfluxDB organization.
                sql_db_username (str): SQL database username.
                sql_db_password (str): SQL database password.
                sql_db_host (str): SQL database host.
                sql_db_port (int): SQL database port.
                sql_db_database_name (str): SQL database name.
                proxies Union[dict[str, str], None]: Proxies.
                sql_db_type (str, optional): SQL Database type engine to use('mysql', 'postgresql' or 'none'), if not defined mysql is used by default.
                    Note: 'none' will result in no SQL engine being created. Hence all functionality that requires a SQL engine will not work.
        """

        self.logger = logging.get_logger(self.__class__.__name__)
        self.influx_organization = config.influx_organization

        # Get db type from config, set 'mysql' if the variable does not exist
        self.sql_db_type = getattr(config, "sql_db_type", "MYSQL")

        if self.sql_db_type not in SupportedSqlTypes.__members__.keys():
            raise ValueError(
                f"Unsupported database sql type '{self.sql_db_type}'. Please use one of the following {SupportedSqlTypes.__members__.keys()}."
            )

        # Set SQL engine according to given sql_db_type
        if self.sql_db_type == SupportedSqlTypes.POSTGRESQL.name:
            self.sql_engine = self._create_postgresql_engine(
                username=config.sql_db_username,
                password=config.sql_db_password,
                host=config.sql_db_host,
                port=config.sql_db_port,
                db=config.sql_db_database_name,
            )
        elif self.sql_db_type == SupportedSqlTypes.NONE.name:
            self.sql_engine = None
        else:
            self.sql_engine = self._create_mysql_engine(
                username=config.sql_db_username,
                password=config.sql_db_password,
                host=config.sql_db_host,
                port=config.sql_db_port,
                db=config.sql_db_database_name,
            )

        self.ktp_api = KtpApi(
            username=config.api_username,
            password=config.api_password,
            admin_username=config.api_admin_username,
            admin_password=config.api_admin_password,
            url=config.api_url,
            proxies=config.proxies,
        )

        self.influx_client = self._create_influx_client(
            token=config.influxdb_token,
            host=config.influxdb_host,
            port=config.influxdb_port,
            organization=config.influx_organization,
        )

        self.influx_query_api = self.influx_client.query_api()
        self.influx_write_api = self.influx_client.write_api(write_options=SYNCHRONOUS)

        # Set geopy proxies
        # https://geopy.readthedocs.io/en/stable/#geopy.geocoders.options
        # https://docs.python.org/3/library/urllib.request.html#urllib.request.ProxyHandler
        # By default the system proxies are respected
        # (e.g. HTTP_PROXY and HTTPS_PROXY env vars or platform-specific proxy settings,
        # such as macOS or Windows native preferences – see
        # urllib.request.ProxyHandler for more details).
        # The proxies value for using system proxies is None.
        geopy.geocoders.options.default_proxies = config.proxies

        _DataInterface._instance = self

    @staticmethod
    def get_instance():
        try:
            return Singleton.get_instance(_DataInterface)
        except KeyError as exc:
            # if _DataInterface not in Singleton._instances:
            raise RuntimeError(
                "No _DataInterface instance initialized. "
                "Please call _DataInterface(config) first."
            ) from exc

    def get_sql_db_type(self):
        return self.sql_db_type

    def _create_influx_client(
        self, token: str, host: str, port: int, organization: str
    ) -> None:
        """Create influx client, namespace-dependend"""
        try:
            return InfluxDBClient(
                url=f"{host}:{port}",
                token=token,
                org=organization,
                timeout=30_000,
            )
        except Exception as exc:
            self.logger("Could not connect to InfluxDB database", exc_info=exc)
            raise

    def _create_mysql_engine(
        self, username: str, password: str, host: str, port: int, db: str
    ):
        """Create MySQL engine.

        Differs from sql_connection in the sense that this write_engine
        *can* write pandas dataframe directly.

        """
        connector = "mysql+mysqlconnector"
        database_url = (
            f"{connector}://{username}:{password}@{host}:{port}/{db}?use_pure=True"
        )
        try:
            return sqlalchemy.create_engine(database_url)
        except Exception as exc:
            self.logger.error("Could not connect to MySQL database", exc_info=exc)
            raise

    def _create_postgresql_engine(
        self, username: str, password: str, host: str, port: int, db: str
    ):
        """Create PostgreSQL engine.

        Differs from sql_connection in the sense that this write_engine
        *can* write pandas dataframe directly.

        """
        connector = "postgresql+psycopg2"
        database_url = f"{connector}://{username}:{password}@{host}:{port}/{db}"
        try:
            return sqlalchemy.create_engine(database_url)
        except Exception as exc:
            self.logger.error("Could not connect to PostgreSQL database", exc_info=exc)
            raise

    def exec_influx_query(self, query: str, bind_params: dict = {}) -> dict:
        """Execute an InfluxDB query.

        When there is data it returns a defaultdict with as key the measurement and
        as value a DataFrame. When there is NO data it returns an empty dictionairy.

        Args:
            query (str): Influx query string.
            bind_params (dict): Binding parameter for parameterized queries

        Returns:
            defaultdict: Query result.
        """
        try:
            return self.influx_query_api.query_data_frame(query)
        except requests.exceptions.ConnectionError as e:
            self.logger.error("Lost connection to InfluxDB database", exc_info=e)
            raise
        except Exception as e:
            self.logger.error(
                "Error occured during executing InfluxDB query", query=query, exc_info=e
            )
            raise

    def exec_influx_write(
        self,
        df: pd.DataFrame,
        database: str,
        measurement: str,
        tag_columns: list,
        organization: str = None,
        field_columns: list = None,
        time_precision: str = "s",
    ) -> bool:
        if field_columns is None:
            field_columns = []
        if type(tag_columns) is not list:
            raise ValueError("'tag_columns' should be a list")
        if organization is None:
            organization = self.influx_organization

        if len(tag_columns) == 0:
            raise ValueError("At least one tag column should be given in 'tag_columns'")

        # Check if a value is nan
        if True in df.isna().values:
            nan_columns = df.columns[df.isna().any()].tolist()
            raise ValueError(
                f"Dataframe contains NaN's. Found NaN's in columns: {nan_columns}"
            )
        # Check if a value is inf
        if df.isin([np.inf, -np.inf]).any().any():
            inf_columns = df.columns[df.isinf().any()].tolist()
            raise ValueError(
                f"Dataframe contains Inf's. Found Inf's in columns: {inf_columns}"
            )

        if True in df.isnull().values:
            nan_columns = df.columns[df.isnull().any()].tolist()
            raise ValueError(
                f"Dataframe contains missing values. Found missing values in columns: {nan_columns}"
            )

        try:
            self.influx_write_api.write(
                record=df,
                data_frame_measurement_name=measurement,
                bucket=f"{database}/autogen",
                data_frame_tag_columns=tag_columns,
                record_field_keys=field_columns,
                write_precision=time_precision,
                org=organization,
            )
            return True
        except Exception as e:
            self.logger.error(
                "Exception occured during writing to InfluxDB", exc_info=e
            )
            raise

    def check_influx_available(self):
        """Check if a basic influx query gives a valid response"""
        query = "buckets()"
        response = self.exec_influx_query(query)
        if isinstance(response, pd.DataFrame):
            available = not response.empty
        else:
            available = False

        return available

    def exec_sql_query(self, query: str, params: dict = None):
        try:
            with self.sql_engine.connect() as connection:
                if params is None:
                    params = {}
                cursor = connection.execute(text(query).bindparams(**params))
                if cursor.cursor is not None:
                    return pd.DataFrame(cursor.fetchall())
        except sqlalchemy.exc.OperationalError as e:
            self.logger.error(
                "Lost connection to {} database".format(self.sql_db_type), exc_info=e
            )
            raise
        except sqlalchemy.exc.ProgrammingError as e:
            self.logger.error(
                "Error occured during executing query", query=query, exc_info=e
            )
            raise
        except sqlalchemy.exc.DatabaseError as e:
            self.logger.error(
                "Can't connect to {} database".format(self.sql_db_type), exc_info=e
            )
            raise

    def exec_sql_write(self, statement: str, params: dict = None) -> None:
        try:
            with self.sql_engine.connect() as connection:
                response = connection.execute(text(statement).bindparams(**params))
                connection.commit()
                self.logger.info(
                    f"Added {response.rowcount} new systems to the systems table in the {self.sql_db_type} database"
                )
        except Exception as e:
            self.logger.error(
                "Error occured during executing query", query=statement, exc_info=e
            )
            raise

    def exec_sql_dataframe_write(
        self, dataframe: pd.DataFrame, table: str, **kwargs
    ) -> None:
        dataframe.to_sql(table, self.sql_engine, **kwargs)

    def check_sql_available(self):
        """Check if a basic SQL query gives a valid response."""
        query = "SELECT 1"

        try:
            response = self.exec_sql_query(query)
            available = response is not None and len(response) > 0

            if available:
                return True
            else:
                print("The SQL query was executed, but no data was returned.")
                return False

        except Exception as e:
            print(f"Error while checking {self.sql_db_type} availability: {e}")
            return False


class SupportedSqlTypes(Enum):
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    NONE = "none"



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
from datetime import datetime, timezone

import numpy as np
import pandas as pd

DEFAULT_FREQ = "15min"
DEFAULT_TZ = timezone.utc


def genereate_datetime_index(start, end, freq=None):
    # Use timezone info from start if given
    if start.tzinfo is None:
        tz = DEFAULT_TZ
    else:
        tz = start.tzinfo

    if not freq:
        freq = DEFAULT_FREQ

    datetime_index = pd.date_range(
        start=start,
        end=end,
        freq=freq,
        tz=tz,
    )
    # Round to given frequency
    datetime_index = datetime_index.round(freq)

    return datetime_index


def process_datetime_range(start, end, freq=None):
    datetime_index = genereate_datetime_index(start, end, freq)

    # Use the index to generate valid start and end times
    datetime_start = datetime_index[0].to_pydatetime()
    datetime_end = datetime_index[-1].to_pydatetime()

    return datetime_start, datetime_end, datetime_index


def parse_influx_result(
    result: pd.DataFrame, aditional_indices: list[str] = None, aggfunc="mean"
) -> pd.DataFrame:
    """Parse resulting DataFrame of flux query to a format we expect in the rest of the lib."""
    indices = ["_time"]
    if aditional_indices is not None:
        indices.extend(aditional_indices)

    result.loc[:, "_time"] = pd.to_datetime(result["_time"])
    result = result.pivot_table(
        columns="_field", values="_value", index=indices, aggfunc=aggfunc
    )
    result = result.reset_index().set_index("_time")
    result.index.name = "datetime"
    result.columns.name = ""
    return result


def _round_down_single_time_diff(time_diff, options):
    """Rounds a time difference to the first smaller option.
    Used to calculate the tAhead for forecasts.

    Options should be sorted"""

    next_smaller_option = next((opt for opt in options if opt < time_diff), None)
    return next_smaller_option


def round_down_time_differences(time_diffs, options):
    """Round a number of time diffs to the first smaller option"""
    sorted_options = np.sort(options)[::-1]
    rounded_times = [
        _round_down_single_time_diff(time_diff, sorted_options)
        for time_diff in time_diffs
    ]
    return rounded_times


def floor_to_closest_time_resolution(time: datetime, resolution_minutes: int):
    return time.replace(
        minute=time.minute - time.minute % resolution_minutes, second=0, microsecond=0
    )



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from openstef_dbc import Singleton
from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.ems import Ems
from openstef_dbc.services.model_input import ModelInput
from openstef_dbc.services.prediction_job import PredictionJobRetriever
from openstef_dbc.services.predictions import Predictions
from openstef_dbc.services.predictor import Predictor
from openstef_dbc.services.splitting import Splitting
from openstef_dbc.services.systems import Systems
from openstef_dbc.services.weather import Weather
from openstef_dbc.services.write import Write


class DataBase(metaclass=Singleton):
    """Provides a high-level interface to various data sources.

    All user/client code should use this class to get or write data. Under the hood
    this class uses various services to interfact with its datasource.
    """

    _instance = None

    # services
    _write = Write()
    _prediction_job = PredictionJobRetriever()
    _weather = Weather()
    _historic_cdb_data_service = Ems()
    _predictor = Predictor()
    _splitting = Splitting()
    _predictions = Predictions()
    _model_input = ModelInput()
    _systems = Systems()

    # write methods
    write_weather_data = _write.write_weather_forecast_data
    write_realised = _write.write_realised
    write_realised_pvdata = _write.write_realised_pvdata
    write_kpi = _write.write_kpi
    write_forecast = _write.write_forecast
    write_apx_market_data = _write.write_apx_market_data
    write_sjv_load_profiles = _write.write_sjv_load_profiles
    write_windturbine_powercurves = _write.write_windturbine_powercurves
    write_energy_splitting_coefficients = _write.write_energy_splitting_coefficients

    # prediction job methods
    get_prediction_jobs_solar = _prediction_job.get_prediction_jobs_solar
    get_prediction_jobs_wind = _prediction_job.get_prediction_jobs_wind
    get_prediction_jobs = _prediction_job.get_prediction_jobs
    get_prediction_job = _prediction_job.get_prediction_job
    get_pids_for_api_key = _prediction_job.get_pids_for_api_key
    get_pids_for_api_keys = _prediction_job.get_pids_for_api_keys
    get_ean_for_pid = _prediction_job.get_ean_for_pid
    get_eans_for_pids = _prediction_job.get_eans_for_pids

    # weather methods
    get_weather_forecast_locations = _weather.get_weather_forecast_locations
    get_weather_data = _weather.get_weather_data
    get_datetime_last_stored_knmi_weatherdata = (
        _weather.get_datetime_last_stored_knmi_weatherdata
    )
    # predictor methods
    get_predictors = _predictor.get_predictors
    get_electricity_price = _predictor.get_day_ahead_electricity_price
    get_load_profiles = _predictor.get_load_profiles
    # historic cdb data service
    get_load_sid = _historic_cdb_data_service.get_load_sid
    get_load_pid = _historic_cdb_data_service.get_load_pid

    # splitting methods
    get_wind_ref = _splitting.get_wind_ref
    get_energy_split_coefs = _splitting.get_energy_split_coefs
    get_input_energy_splitting = _splitting.get_input_energy_splitting
    # predictions methods
    get_predicted_load = _predictions.get_predicted_load
    get_predicted_load_tahead = _predictions.get_predicted_load_tahead
    get_prediction_including_components = (
        _predictions.get_prediction_including_components
    )
    get_forecast_quality = _predictions.get_forecast_quality
    # model input methods
    get_model_input = _model_input.get_model_input
    get_wind_input = _model_input.get_wind_input
    get_power_curve = _model_input.get_power_curve
    get_solar_input = _model_input.get_solar_input
    # systems methods
    get_systems_near_location = _systems.get_systems_near_location
    get_systems_by_pid = _systems.get_systems_by_pid
    get_pv_systems_with_incorrect_location = (
        _systems.get_pv_systems_with_incorrect_location
    )
    get_random_pv_systems = _systems.get_random_pv_systems
    get_api_key_for_system = _systems.get_api_key_for_system
    get_api_keys_for_systems = _systems.get_api_keys_for_systems

    def __init__(self, config):
        """Construct the DataBase singleton.

        Initialize the datainterface and api. WARNING: this is a singleton class when
        calling multiple times with a config argument no new configuration will be
        applied.

        Args:
            config: Configuration object. with the following attributes:
                api_username (str): API username.
                api_password (str): API password.
                api_admin_username (str): API admin username.
                api_admin_password (str): API admin password.
                api_url (str): API url.
                influxdb_token (str): Token to authenticate to InfluxDB.
                influxdb_host (str): InfluxDB host.
                influxdb_port (int): InfluxDB port.
                influx_organization (str): InfluxDB organization.
                sql_db_username (str): SQL database username.
                sql_db_password (str): SQL database password.
                sql_db_host (str): SQL database host.
                sql_db_port (int): SQL database port.
                sql_db_database_name (str): SQL database name.
                proxies Union[dict[str, str], None]: Proxies.
                sql_db_type (str, optional): SQL Database type ('mysql' or 'postgresql').
        """

        self._datainterface = _DataInterface(config)
        # Ktp api
        self.ktp_api = self._datainterface.ktp_api

        DataBase._instance = self



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

import warnings
from json.decoder import JSONDecodeError

import pandas as pd
import pytz
import requests
import time

from openstef_dbc.log import logging


class ApiException(Exception):
    pass


class KtpApi:
    """Used for all api traffic to Influx API"""

    __INFLUX_TEST_URL = "api/_core/health/liveness"
    __MEASUREMENTS_URL = "api/measurements"
    __SYSTEMS_URL = "admin/systems"
    __TRACY_JOB_URL = "admin/tracyjobs"

    __ALL_TRACY_JOBS = -999

    def __init__(
        self,
        username,
        password,
        url,
        admin_username=None,
        admin_password=None,
        proxies=None,
        timeout=None,
    ):
        """Define api class"""
        # get config
        self.logger = logging.get_logger(self.__class__.__name__)

        self.base_url = url
        self.credentials = username, password
        self.admin_credentials = admin_username, admin_password
        self.proxies = proxies

        self.timeout = timeout
        # Test if the connection works
        self.test()

    def _get_credentials(self, path):
        """
        Returns the admin or api credentials, depending on the path

        :param path: URL to access
        :return: admin or api credentials
        """
        if "admin" in path:
            return self.admin_credentials
        else:
            return self.credentials

    def _get(self, path, params=None):
        """Preform a GET request on the API.

        Args:
            path: the URL to get the data from
            params: the query params.

        Raises: Warning if the response status != 200

        Returns:
            requests.models.Response object
        """

        endpoint = requests.compat.urljoin(self.base_url, path)
        credentials = self._get_credentials(path)

        r = requests.get(
            endpoint,
            auth=credentials,
            proxies=self.proxies,
            params=params,
            timeout=self.timeout,
        )
        if r.status_code != 200:
            warnings.warn(
                "GET for url {} failed with status {} and message: {}".format(
                    endpoint, r.status_code, r.text
                )
            )

        return r

    def _post(self, path, payload, params=None):
        """
        Issue a POST to the given path with the given payload
        :param path: URL to post to
        :param payload: the POST payload
        :return: `Response <Response>` object
        """
        # TODO: Why not have _post raise an exception when it fails
        # (Note: need to change code which checks for the success string)
        endpoint = requests.compat.urljoin(self.base_url, path)
        credentials = self._get_credentials(path)

        p = requests.post(
            endpoint,
            json=payload,
            auth=credentials,
            proxies=self.proxies,
            params=params,
            timeout=self.timeout,
        )
        return p

    def _patch(self, path, payload):
        """
        Issue a PATCH request to the given path with the given payload

        :param path:  URL to PATCH the payload to
        :param payload: payload with fields to update
        :return: updated element
        """

        endpoint = requests.compat.urljoin(self.base_url, path)
        credentials = self._get_credentials(path)

        p = requests.patch(
            endpoint,
            json=payload,
            auth=credentials,
            proxies=self.proxies,
            timeout=self.timeout,
        )

        if p.status_code != 200:
            warnings.warn(
                "PATCH for url {} with payload {} failed: {}".format(
                    endpoint, payload, p.text
                )
            )

        return p.json()

    def _put(self, path, payload):
        endpoint = requests.compat.urljoin(self.base_url, path)
        credentials = self._get_credentials(path)

        u = requests.put(
            endpoint,
            json=payload,
            auth=credentials,
            proxies=self.proxies,
            timeout=self.timeout,
        )

        if u.status_code != 200:
            warnings.warn(
                "PUT for url {} with payload {} failed: {}".format(
                    endpoint, payload, u.text
                )
            )

        return u.json()

    def _delete(self, path):
        endpoint = requests.compat.urljoin(self.base_url, path)
        credentials = self._get_credentials(path)

        d = requests.delete(
            endpoint, auth=credentials, proxies=self.proxies, timeout=self.timeout
        )
        if d.status_code != 204:
            warnings.warn("DELETE for url {} failed: {}".format(endpoint, d.text))

    def test(self):
        """Test connection by pinging the influx database"""
        try:
            r = self._get(KtpApi.__INFLUX_TEST_URL)
        except Exception as e:
            warnings.warn(
                f"Could not establish connection with ktp-api. Continuing anyway... {e}"
            )
            r = dict(status=200)
        return r

    def get_customers(self):
        """
        :return: dataframe with all customers
        """
        r = self._get("admin/customers")

        customers = pd.DataFrame(r.json())
        return customers

    # NOTE this is the new implementation
    def post_measurement(self, measurement, api_key):
        measurements_url = KtpApi.__MEASUREMENTS_URL

        payload = measurement.to_dict()

        params = {"key": api_key}

        try:
            r = self._post(measurements_url, payload, params)
        except Exception as e:
            raise self._build_api_exception_from_requests_exception(e)

        if r.status_code != 201:
            raise self._build_api_exception_from_response(r)

        response = r.json()

        return response

    # NOTE this one will be combined and depreciated in the future
    def post_measurements(self, data, sid, api_key):
        """
        This function can be used to post measured data to our Java API.

        :param data: pd.DataFrame(index = datetime, columns = [output])
        :param sid: str SystemId
        :param api_key: the API key
        :return: string with either a message, or the return value of the POST request
        """

        # API set-up
        measurement_url = KtpApi.__MEASUREMENTS_URL + "/?key={0}".format(api_key)

        ####
        # Change format of data
        ds = data.reset_index()
        # Check if a timezone is present, else add utc
        if ds["datetime"][0].tzinfo is None:
            self.logger.warning("Adding timezone, assuming UTC")
            ds["datetime"] = [x.replace(tzinfo=pytz.UTC) for x in ds["datetime"]]
        ds.set_index("datetime").tz_convert("CET").reset_index()
        # Convert time to isoformat
        ds["datetime"] = ds["datetime"].apply(
            lambda x: x.tz_convert("CET")
            .isoformat()
            .replace("+00:00", "+0000")
            .replace("+01:00", "+0100")
            .replace("+02:00", "+0200")
        )

        payload = dict(sid=sid, data=ds.to_dict("records"))

        p = self._post(measurement_url, payload)

        if p.ok:
            message = "Post successful! Entries: " + str(len(p.json()["data"]))
        else:
            message = p.json()
        return message

    def get_system(self, sid):
        systems_url = f"{KtpApi.__SYSTEMS_URL}/{sid}"
        params = {}

        try:
            r = self._get(systems_url, params)
        except Exception as e:
            raise self._build_api_exception_from_requests_exception(e)

        if r.status_code != 200:
            raise self._build_api_exception_from_response(r)

        system = r.json()

        return system

    def get_systems(self, origin=None, limit=None):
        """Get systems.

        Args:
            origin (str, optional): System origin. Defaults to None.
            limit (int, optional): Limit the number of systems. Defaults to None.

        Raises:
            ApiException

        Returns:
            list: Raw API response systems, list of dicts.
        """
        systems_url = KtpApi.__SYSTEMS_URL
        params = {}

        if origin is not None:
            params["origin"] = origin.upper()
        if limit is not None:
            params["limit"] = limit

        try:
            r = self._get(systems_url, params)
        except Exception as e:
            raise self._build_api_exception_from_requests_exception(e)

        if r.status_code != 200:
            raise self._build_api_exception_from_response(r)

        systems = r.json()

        return systems

    def update_system(self, sid, fields):
        """
        Updates the fields for the given system

        :param sid: system ID
        :param fields:  fields to update
        :return: updated system
        """
        systems_url = KtpApi.__SYSTEMS_URL + "/{0}".format(sid)
        return self._patch(systems_url, fields)

    def _tracy_function_tolower(self, jobs):
        if isinstance(jobs, list):
            for job in jobs:
                job["function"] = job["function"].lower()
        else:
            jobs["function"] = jobs["function"].lower()
        return jobs

    def get_all_tracy_jobs(self, inprogress=__ALL_TRACY_JOBS):
        """
        Retrieves all Tracy Jobs from the Influx API. If inprogress is 0, all entries from the database will be returned
        that either have inprogress to 0 or to NULL

        :param inprogress: filter on inprogress. If no filter is given, all Tracy Jobs are returned.
                           inprogress must be between 0 and 2
        :return: a list with all Tracy Jobs
        """
        r = self._get(KtpApi.__TRACY_JOB_URL)

        # add a quick and dirty retry.
        max_retries = 3
        while max_retries & (r.status_code != 200):
            r = self._get(KtpApi.__TRACY_JOB_URL)
            time.sleep(2)

        # Finally
        if r.status_code != 200:
            # Raise exception of no valid response.
            # Alternatively, we could chose to return empty job list.
            # Lets try this first
            raise ConnectionError(
                "API endpoint to retrieve tracy jobs was not available"
            )
        jobs = self._tracy_function_tolower(r.json())

        # No filter
        if inprogress == KtpApi.__ALL_TRACY_JOBS:
            return jobs

        return list(filter(lambda x: (x.get("inprogress", 0) == inprogress), jobs))

    def get_tracy_job(self, pid, function="train_model"):
        """
        Return the tracy job from the database

        :param pid:
        :param function:
        :return: return the tracy job if it exsits or None if it doesn't
        """
        jobs = self.get_all_tracy_jobs()

        try:
            return next(
                item
                for item in jobs
                if item["args"] == str(pid) and item["function"].lower() == function
            )
        except StopIteration:
            pass

        return None

    def add_tracy_job(self, pid, function="train_model"):
        """
        Adds the function to the tracy todo list

        :param pid: args of the job
        :param function: function to execute
        :return: The Tracy Job
        """
        r = self._post(
            KtpApi.__TRACY_JOB_URL, {"args": str(pid), "function": function.upper()}
        )

        # Check status code. 200 is OK. 400 and 409 might be returned if the job already exists which is also fine
        if r.status_code == 200:
            return self._tracy_function_tolower(r.json())
        elif r.status_code == 409:
            # Already exists. Return existing job
            return self.get_tracy_job(pid, function)
        else:
            warnings.warn(
                "Unable to add Tracy Job. Status code: {0}. Message: {1}".format(
                    r.status_code, r.text
                )
            )

    def update_tracy_job(self, job):
        """
        Update the given Tracy Job by setting inprogess to the requested value

        :param job: prediction job to update
        :return: updated tracy job
        """

        job["function"] = job["function"].upper()
        self._put(KtpApi.__TRACY_JOB_URL + "/{0}".format(job["id"]), job)
        job = self._tracy_function_tolower(job)
        return job

    def delete_tracy_job(self, job):
        """
        Deletes the tracy job

        :param job: job to delete
        :return: -
        """
        return self._delete(KtpApi.__TRACY_JOB_URL + "/{0}".format(job["id"]))

    def _build_api_exception_from_response(self, response):
        code = response.status_code
        method = response.request.method
        path_url = response.request.path_url
        # API specific info
        try:
            r = response.json()
            api_error = r["error"]
            api_message = r["message"]
            error_description = f"error: {api_error}, message: {api_message}"
        except JSONDecodeError:
            error_description = response.text

        return ApiException(
            f"API non 200 status code: [{code}] [{method}] [{path_url}]. "
            f"{error_description}"
        )

    def _build_api_exception_from_requests_exception(self, e):
        return ApiException(f"An exception occured while making an API requests: {e}")



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

import warnings
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("openstef_dbc")
except PackageNotFoundError:
    # package is not installed
    pass


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        elif cls in cls._instances and len(args) > 0 or len(kwargs) > 0:
            warnings.warn(
                f"Singleton class '{cls.__name__}'' already initialized, arguments are ignored"
            )
        return cls._instances[cls]

    def clear(cls):
        cls._instances = {}

    @classmethod
    def get_instance(cls, instance_cls):
        return cls._instances[instance_cls]



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Union, List
import datetime
import re

from datetime import timedelta

import pandas as pd
import structlog
from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.systems import Systems
from openstef_dbc.utils import process_datetime_range


class Ems:
    def __init__(self) -> None:
        self.logger = structlog.get_logger(self.__class__.__name__)

    def get_load_sid(
        self,
        sid: Union[List[str], str],
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str,
        aggregated: bool = True,
        average_output: bool = False,
        include_n_entries_column=False,
    ) -> pd.DataFrame:
        """Get the load for a single or multiple system id's.

            Get Measurements for given sid or list of sids. If no result is found,
            return empty dataframe.

        Args:
            sid(Union[str, List[str]]): System id's
            datetime_start (datetime): Start datetime.
            datetime_end (datetime): End datetime.
            forecast_resolution (str): The forecast resolution, for example '15min'
            aggregated (boolean): Should the results be aggregated per sid or not.
            average_output (boolean):  Should the agregation be an average or not.
            include_n_entries_column (boolean): Should the number of systems agregated be returned or not.

        Returns:
            - pd.DataFrame(index=datetimeIndex, columns)
            if aggregated: columns = [load, nEntries], else: columns = sid"""

        # Process datetimes (rounded, timezone, frequency) and generate index
        datetime_start, datetime_end, _ = process_datetime_range(
            start=datetime_start,
            end=datetime_end,
            freq=forecast_resolution,
        )
        # Define empty bind_params dict for query parametrization
        bind_params = {}

        # Convert sid to list
        if type(sid) is str:
            sid = [sid]

        # Prepare sid query string
        if len(sid) == 1:
            bind_params[f"param_sid_0"] = sid[0]
            sidsection = f'r.system == "{bind_params["param_sid_0"]}"'
        else:
            # Create parameters for each sid
            for i in range(len(sid)):
                bind_params[f"param_sid_{i}"] = sid[i]

            # Build parameters in query
            section = " or r.system == ".join(
                ['"' + s + '"' for s in list(bind_params.values())]
            )
            sidsection = f"r.system == {section}"

        bind_params.update(
            {
                "dstart": datetime_start.isoformat(),
                "dend": datetime_end.isoformat(),
            }
        )

        # Prepare query
        if aggregated:
            forecast_resolution_timedelta = timedelta(
                minutes=int(pd.Timedelta(forecast_resolution).total_seconds() / 60)
            )
            # Extending the range is necessary to make sure the final timestamps are also in
            # the reponse after aggregations.
            bind_params["dend"] = (
                datetime_end + 2 * forecast_resolution_timedelta
            ).isoformat()

            forecast_resolution_influx_format = forecast_resolution.replace(
                "T", "m"
            ).replace("min", "m")
            query = f"""
                data = from(bucket: "realised/autogen") 
                    |> range(start: {bind_params['dstart']}, stop: {bind_params['dend']}) 
                    |> filter(fn: (r) => r._measurement == "power")
                    |> filter(fn: (r) => r._field == "output")
                    |> filter(fn: (r) => {sidsection})
                    |> aggregateWindow(every: {forecast_resolution_influx_format}, fn: mean)

                data
                    |> group() |> aggregateWindow(every: {forecast_resolution_influx_format}, fn: sum)
                    |> yield(name: "load")

                data
                    |> group() |> aggregateWindow(every: {forecast_resolution_influx_format}, fn: count)
                    |> yield(name: "nEntries")
            """
        else:
            query = f"""
                from(bucket: "realised/autogen") 
                    |> range(start: {bind_params['dstart']}, stop: {bind_params['dend']}) 
                    |> filter(fn: (r) => r._measurement == "power")
                    |> filter(fn: (r) => r._field == "output")
                    |> filter(fn: (r) => {sidsection})
                    |> pivot(rowKey:["_time"], columnKey: ["system"], valueColumn: "_value")
            """

        # Query load
        result_raw = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        if isinstance(result_raw, pd.DataFrame) and result_raw.empty:
            self.logger.warning(
                f"Probably no load data avalailable for the following system(s): \n {sid} \n"
                f"The following query yields an empty dataframe: {query}"
            )
            return pd.DataFrame()

        if aggregated:
            # `result_raw`, contains the outcome of a flux query that returns a list of two dataframes.
            # The order of the two dataframes within the list can differ from one run to the next.
            # Therefore the renaming of the `_value` column is done dynamically.
            result = pd.concat(
                [
                    result_raw[0][["_time", "_value"]].rename(
                        columns={"_value": result_raw[0]["result"][0]}
                    ),
                    result_raw[1][["_value"]].rename(
                        columns={"_value": result_raw[1]["result"][0]}
                    ),
                ],
                axis=1,
            ).set_index("_time")

            # Shifting the index two timeperiods back, because flux takes the right instead
            # of the left boundary when aggregating over a time window. In the flux query,
            # two aggregations are performed over a `forecast_resolution`-sized time window.
            result = result.shift(periods=-2, freq=forecast_resolution)
            # The first two rows are dropped, since their timestamps are before the `datetime_start`.
            result = result.iloc[2:, :]

            result = result.dropna()
            if average_output:
                result["load"] = result["load"] / result["nEntries"]

            outputcols = ["load"]

            if include_n_entries_column:
                outputcols.append("nEntries")

            return result[outputcols]
        else:
            result = result_raw.set_index("_time")[sid]
            result = result.resample(forecast_resolution).mean()
            return result

    def get_load_pid(
        self,
        pid: int,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str = "15min",
        aggregated: bool = True,
        ignore_factor: bool = False,
    ) -> pd.DataFrame:
        """Get load(s) for a given prediction job id.

        Retrieve the load for all systems which belong to a given prediction job id. The
        loads will always be multiplied by both the `polarity` as well as the `factor`
        of any given system. Except when `ignore_factor` is set to True, in that case
        the load will not be multiplied  with the `factor`.

        Aggregated will return a single column dataframe. While non aggregated
        will return a multi column dataframe (one column per system).

        NOTE: if the polarity is set to `0` a positive polarity is assumed.

        Args:
            pid (int): id of the prediction job
            datetime_start (str): Datetime start range
            datetime_end (str): Datetime end range
            forecast_resolution: timeresolution of result in pd.resample() format
            aggregated (bool): Should the result for multiple sid's be aggregated?
                Default: True
            ignore_factor (bool): When set to True, the `factor` will not be applied.
                Default: False

        Returns:
            (pd.DataFrame): Load
        """

        # Process datetimes (rounded, timezone, frequency) and generate index
        datetime_start, datetime_end, _ = process_datetime_range(
            start=datetime_start,
            end=datetime_end,
            freq=forecast_resolution,
        )

        # Use optimized load retrieval if possible
        if aggregated and not ignore_factor:
            return self._get_load_pid_optimized(
                pid, datetime_start, datetime_end, forecast_resolution
            )

        # Get systems that belong to this prediction
        systems = Systems().get_systems_by_pid(pid)

        if systems.empty:
            raise ValueError(f"No systems found for given pid ({pid}).")

        # obtain load for all systems
        systems_load = self.get_load_sid(
            list(systems.system_id),
            datetime_start,
            datetime_end,
            forecast_resolution,
            aggregated=False,
        )

        # if load is empty, raise a warning and return empty dataframe
        if len(systems_load) == 0:
            self.logger.warning("No load data retrieved. Returning empty dataframe")
            return pd.DataFrame()

        # Check if all requested systems have a historic load, otherwise, give a warning
        # and ignore 'missing' systems
        missing_systems = [
            x for x in systems.system_id if x not in systems_load.columns
        ]
        num_missing_systems = len(missing_systems)
        if num_missing_systems > 0:
            msg = (
                f"There is/are {num_missing_systems} ({', '.join(missing_systems)}) "
                f"system(s) without a load for pid {pid}, ignore this/these system(s)."
            )
            self.logger.warning(
                msg,
                pid=pid,
                num_missing_systems=num_missing_systems,
                missing_systems=missing_systems,
            )
            systems = systems[~systems.system_id.isin(missing_systems)]

        # Apply positive or negative polariy to systems
        for system_id in systems.system_id:
            polarity = systems.loc[systems.system_id == system_id].polarity.iloc[0]
            if polarity == 0:
                self.logger.warning(
                    "Polarity not set use 1 by default", system_id=system_id
                )
                polarity = 1
            systems_load[system_id] *= polarity

        if ignore_factor is False:
            # apply predictions_systems factor -> should we add or subtract this load
            for system_id in systems.system_id:
                factor = systems.loc[systems.system_id == system_id].factor.iloc[0]
                systems_load[system_id] *= factor

        # if aggregated is False return non aggregated dataframe
        if aggregated is False:
            return systems_load

        # if aggregated is true all normal and load correction systems will be added
        total_load = pd.DataFrame()

        # aggregrate the load by adding all system loads
        total_load["load"] = systems_load.sum(axis=1)

        return total_load

    def _get_load_pid_optimized(
        self,
        pid: int,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str = "15min",
    ) -> pd.DataFrame:
        """Gets load data for a pid.
        This method optimizes the way it retrieves data and is therefore less flexible as get_load_pid.
        It is however much faster for prediction jobs with a large amount of sid's.

        Args:
            pid (int): id of the prediction job
            datetime_start (str): Datetime start range
            datetime_end (str): Datetime end range
            forecast_resolution: timeresolution of result in pd.resample() format

        Returns:
            (pd.DataFrame): Load

        """
        # Get systems that belong to this prediction
        systems = Systems().get_systems_by_pid(pid)

        # Set factor or polarity of 0 to 1
        systems.loc[systems["factor"] == 0, "factor"] = 1
        systems.loc[systems["polarity"] == 0, "polarity"] = 1
        systems["effective_factor"] = systems["factor"] * systems["polarity"]

        # Build dict with unique effective factors as keys and lists of corresponding systems as values
        effective_factors = {
            effective_factor: systems[systems["effective_factor"] == effective_factor]
            for effective_factor in set(systems["effective_factor"])
        }

        combined_load = pd.DataFrame()
        # Retrieve load for each unique effective_factor
        for effective_factor, sids in effective_factors.items():
            load = (
                self.get_load_sid(
                    sids.system_id.to_list(),
                    datetime_start,
                    datetime_end,
                    forecast_resolution,
                    aggregated=True,
                )
                * effective_factor
            )

            # Combine individual results
            load = load.rename(columns=dict(load=effective_factor))

            # Use merge so potential gaps in the individual timeseries do not cause issues
            combined_load = combined_load.merge(
                load, left_index=True, right_index=True, how="outer"
            )

        #  Return sum all load columns
        return pd.DataFrame(combined_load.sum(axis=1).rename("load"))

    @staticmethod
    def _check_influx_group_by_time_statement(statement: str) -> None:
        # Validate forecast resolution to prevent injections
        if not re.match(
            r"[0-9]+([unµm]?s|m|h|d|w)",
            statement,
        ):
            raise ValueError("Forecast resolution does not have the allowed format!")



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Union, Tuple, List
from datetime import datetime, timedelta

import geopy
import geopy.distance
import numpy as np
import pandas as pd
import pytz
import structlog
import warnings
from influxdb_client.client.warnings import MissingPivotFunction

from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.write import Write
from openstef_dbc.settings import Settings
from openstef_dbc.utils import genereate_datetime_index, parse_influx_result

warnings.simplefilter("ignore", MissingPivotFunction)


class Weather:
    def __init__(self) -> None:
        self.logger = structlog.get_logger(self.__class__.__name__)

    def get_weather_forecast_locations(
        self, country: str = "NL", active: int = 1
    ) -> List[dict]:
        """Get weather forecast locations.

        Returns:
            List[dict]: List of locations with keys:
                ["city"]:       City of the weather forecast.
                ["lat"]:        Lattitude coordinate.
                ["lon"]:        Longitude coordinate.
                ["country"]:    Country code (2-letter: ISO 3166-1)
        """

        bind_params = {"country": country, "active": active}

        query = """
            SELECT input_city as city, lat, lon, country
            FROM weatherforecastlocations
            WHERE country = :country AND active = :active
        """
        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)

        locations = result.to_dict("records")

        return locations

    def _get_nearest_weather_locations(
        self,
        location: Union[Tuple[float, float], str],
        country: str = "NL",
        threshold: float = 150.0,
        number_locations: int = 1,
    ) -> pd.Series:
        """Find the nearest weather forecast locations.

        Function that, given an location, finds the nearest locations for which a
        weatherforecast is available. A warning is generated when the distance is
        greater than a certain distance (threshold).

        If multiple weather locations have the same distance, only one is returned.

        Args:
            location (str, tuple): Name of the location/city or coordinates (lat, lon).
            country (str): Country code (2-letter: ISO 3166-1).
            threshold (int): Maximum distance [km] before a warning is generated.
            number_locations (int): number of weather locations desired

        Returns:
            Pd.Series: Name of the weather location(s) (distance as index)
        """
        # Get all available cities
        weather_locations = self.get_weather_forecast_locations(
            country=country, active=1
        )

        # If location is string, convert to (lat, lon)
        if type(location) is str:
            location_coordinates = self._get_coordinates_of_location(location)
        else:
            location_coordinates = location

        # Now we have the coordinates of the input_city. Next, find nearest weather location
        distance_dfs = []
        for weather_location in weather_locations:
            coordinates = (weather_location["lat"], weather_location["lon"])

            if np.nan in coordinates:
                raise ValueError(
                    "No coordinates found, pleas make sure a prope location "
                    "is configured in the mySQL database"
                )

            distance = round(
                geopy.distance.geodesic(coordinates, location_coordinates).km, 1
            )
            city = weather_location["city"]
            distance_df = pd.DataFrame([{"distance": distance, "input_city": city}])
            distance_dfs.append(distance_df)

        distances = pd.concat(distance_dfs).set_index("distance")

        nearest_location = distances["input_city"].sort_index().iloc[0:number_locations]

        # Find closest weather location
        if distances.index.min() < threshold:
            return nearest_location

        raise Warning(
            "Closest weather location is farther than threshold: {dist} > {threshold}".format(
                dist=str(min(distances)), threshold=str(threshold)
            )
        )

    def _get_coordinates_of_location(self, location_name: str) -> Tuple[float, float]:
        """Get lat, lon coordinates of location.

        The function tries to get the coordinates from our own sql database,
        if the name is not yet present, Nominatim() is used to retrieve
        lat lons. It then proceeds to add the values to database.

        Args:
            input_city(str): Name of the location/city.

        Returns:
            (tuple): Coordinated:
                [0]: Lattitude
                [1]: Longitude
        """

        # Query corresponding (lat, lon) from SQL database
        binding_params = {"city": location_name}
        query = "SELECT lat, lon from NameToLatLon where regionInput = :city"
        location = _DataInterface.get_instance().exec_sql_query(query, binding_params)

        # If not found
        if len(location) == 0:
            # Get (lat, lon) of location (via google)
            location = geopy.geocoders.Nominatim().geocode(location_name)
            location = (location.latitude, location.longitude)

            # Write found location to MySQL database
            # NOTE Not sure if we should want this. perhaps better to fail and do
            # manual write
            Write().write_location(location_name, location)
        else:
            location = (location["lat"][0], location["lon"][0])

        return location

    def _get_source_run(
        self, forecast_datetime: pd.Series, tAhead: pd.Series
    ) -> pd.Series:
        """Compute the datetime when weather forecast was created

        Args:
            forecast_datetime (pd.Series[datetime]): forecasted datetime.
            tAhead (pd.Series[(int, float)]: forecasting horizon in hours

        Retuns :
            pd.Series of new datetimes
        """

        if not pd.api.types.is_datetime64_any_dtype(forecast_datetime):
            raise ValueError("forecast_datetime must be a Series of datetime.")

        if not pd.api.types.is_numeric_dtype(tAhead):
            raise ValueError("tahead must be a Series of int or float.")

        if len(forecast_datetime) != len(tAhead):
            raise ValueError("forecast_datetime and tAhead must have the same length.")

        # Compute new datetimes
        return forecast_datetime - pd.to_timedelta(tAhead, unit="h")

    def _combine_weather_sources(
        self, result: pd.DataFrame, source_order: List = None
    ) -> pd.DataFrame:
        """
        Function that complete's the weatherdata from the influx database retrieved using get_weatherdata
        Additionally, weatherdata from several sources is combined to a single forecast.
        The preferred source_order is given as argument.

        Args:
            result (pd.DataFrame): return value from the function 'get_weatherdata'
            source_order (list): which weather models should be used (first).
                Options: "OWM", "DSN", "WUN", "harmonie", "harm_arome", "harm_arome_fallback", "optimum", "icon"
                Default: '"harm_arome", "harm_arome_fallback", "harmonie", "icon", "DSN"'.
                This combines harmonie, harm_arome, harm_arome_fallback, ICON and DSN,
                taking the (heuristicly) best available source for each moment in time

        Returns:
            pd.DataFrame: The most recent weather prediction of several sources combined into one dataframe

        Example:
            client = influxdb.DataFrameClient()
            df = QueryWeatherData(client, location='Volkel',
                                  weatherParams=["clouds", "radiation"], source='optimum')
            print(df.head())
        """
        # Combine the multiple sources in the optimal way
        ####
        # step 1: Create list of multiple dataframes,
        # check which of the 'optimum' sources are actually in the list
        if source_order is None:
            source_order = Settings.optimum_weather_sources

        active_sources = [s for s in source_order if s in set(result.source)]
        # for each source a seperate dataframe
        df_list = [result[result.source == s] for s in active_sources]

        # Step two, combine the individual dataframe,
        # keeping data from the most-preferred datasource
        complete_df = df_list.pop(0)
        while len(df_list) > 0:
            df_new_source = df_list.pop(0)
            complete_df = complete_df.combine_first(df_new_source)

        return complete_df

    def get_weather_data(
        self,
        location: Union[Tuple[float, float], str],
        weatherparams: List[str],
        datetime_start: datetime = datetime.utcnow() - timedelta(days=14),
        datetime_end: datetime = datetime.utcnow() + timedelta(days=2),
        source: Union[List[str], str] = "optimum",
        resolution: str = "15min",
        country: str = "NL",
        number_locations: int = 1,
        type: str = "smallest_tAhead",
    ) -> pd.DataFrame:
        """Get weather data from database.

        Additionally, weatherdata from several sources is combined to a single forecast.
        When the source equals "optimum", data from harmonie is preferred,
            completed by data from harm_arome and DSN

        Args:
            location (str): Location. Should be in the stored locations. e.g. Volkel
            weatherparams  (str or list of str): weatherparams that are required.
                Params include: ["clouds", "radiation", "temp", "winddeg", "windspeed", "windspeed_100m", "pressure",
                "humidity", "rain", 'mxlD', 'snowDepth', 'clearSky_ulf', 'clearSky_dlf', 'ssrunoff']
            datetime_start (datetime) : start date time. Default: 14 days ago
            datetime_end (datetime): end date time. Default: now + 2 days.
            source (str or list of str): which weather models should be used.
                Options: "OWM", "DSN", "WUN", "harmonie", "harm_arome", "harm_arome_fallback", "icon", "optimum",
                Default: 'optimum'. This combines harmonie, harm_arome, icon and DSN,
                taking the (heuristicly) best available source for each moment in time
            resolution (str): Time resolution of the returned data, default: "15min"
            country (str): Country code (2-letter: ISO 3166-1). e.g. NL
            number_locations (int): number of weather locations desired
            type (str) : type of weather forecast (smallest_tAhead of multiple_tAheads)
        Returns:
            pd.DataFrame: The most recent weather prediction

        Example:
            client = influxdb.DataFrameClient()
            df = QueryWeatherData(client, location='Volkel',
                                  weatherParams=["clouds", "radiation"], source='optimum')
            print(df.head())
        """

        # Converting to datetime objet with tz attribute
        datetime_start = pd.to_datetime(datetime_start)
        datetime_end = pd.to_datetime(datetime_end)

        # Convert to UTC and remove UTC as note
        if datetime_start.tz is not None:
            datetime_start = datetime_start.tz_convert("UTC").tz_localize(None)

        if datetime_end.tz is not None:
            datetime_end = datetime_end.tz_convert("UTC").tz_localize(None)

        location_name = self._get_nearest_weather_locations(
            location=location, country=country, number_locations=number_locations
        )

        # Make a list of the source and weatherparams.
        # Like this, it also works if source is a string instead of multiple values
        if isinstance(source, str):
            source = [source]
        if isinstance(weatherparams, str):
            weatherparams = [weatherparams]

        # Try to get the data from influx.
        if "optimum" in source:
            # so the query return all
            source = Settings.optimum_weather_sources
            combine_sources = True
        else:
            combine_sources = False

        # Initialize binding params
        bind_params = {
            "_start": datetime_start,
            "_stop": datetime_end,
        }

        # Initialise strings for the querying influx, it is not possible to parameterize this string
        weather_params_str = '" or r._field == "'.join(weatherparams)
        weather_models_str = '" or r.source == "'.join(source)
        weather_location_name_str = '" or r.input_city == "'.join(
            location_name.to_list()
        )

        if type == "smallest_tAhead":
            weather_measurement_str = "weather"
            influx_indices = ["source", "input_city"]
            grouping_indices = ["source", "input_city"]
        elif type == "multiple_tAheads":
            weather_measurement_str = "weather_tAhead"
            influx_indices = ["source", "input_city", "tAhead"]
            grouping_indices = ["source", "input_city", "created"]

        # Create the query
        query = f"""
            from(bucket: "forecast_latest/autogen") 
                |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                |> filter(fn: (r) => r._measurement == "{weather_measurement_str}" and (r._field == "{weather_params_str}") and (r.source == "{weather_models_str}") and (r.input_city == "{weather_location_name_str}"))
        """

        # Execute Query
        result = _DataInterface.get_instance().exec_influx_query(query)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[
                ["_value", "_field", "_time", "source", "input_city"]
            ]

        # Check if response is empty
        if not result.empty:
            result = parse_influx_result(result, influx_indices)
        else:
            self.logger.warning("No weatherdata found. Returning empty dataframe")
            return pd.DataFrame(
                index=genereate_datetime_index(
                    start=datetime_start,
                    end=datetime_end,
                    freq=resolution,
                )
            )

        # Create a single dataframe with combined sources
        if combine_sources:
            self.logger.info("Combining sources into single dataframe")
            result = self._combine_weather_sources(result)
            result["source"] = "optimum"

        # Compute source_run
        if type == "multiple_tAheads":
            result["created"] = self._get_source_run(result.index, result.tAhead)

        # Interpolate if nescesarry by input_city, source (and tAhead)
        with pd.option_context("future.no_silent_downcasting", True):
            result = (
                result.groupby(grouping_indices)
                .resample(resolution, include_groups=False)
                .interpolate(limit=11)
                .reset_index(grouping_indices)
            )

        if number_locations == 1:
            result = result.drop(columns="input_city")
        else:
            # Adding distance information en km
            result = (
                result.reset_index()
                .merge(location_name.reset_index(), how="left", on="input_city")
                .set_index("datetime")
            )

        return result

    def get_datetime_last_stored_knmi_weatherdata(
        self, source: str = "harm_arome"
    ) -> datetime:
        """Returns datetime of latest KNMI run in influx Database. If no run is found return first unix time."""
        query = f"""
            from(bucket: "forecast_latest/autogen" )   
                |> range(start: - 2d)
                |> filter(fn: (r) => r._measurement == "weather" and r.source == "{source}" and r._field == "source_run")
                |> max()
        """
        result = _DataInterface.get_instance().exec_influx_query(query)

        if isinstance(result, pd.DataFrame) and not result.empty:
            # Get latest run
            latest_unix = result["_value"].max()
        else:
            latest_unix = 0  # Return first unix time if no run is found

        # unix timestamp of most recent stored weather forecast created
        last_stored_run = datetime.fromtimestamp(latest_unix, pytz.utc)
        return last_stored_run



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from dataclasses import dataclass, asdict
from typing import List
from datetime import datetime

DATETIME_FMT = "%Y-%m-%dT%H:%M:%S%z"


@dataclass
class MeasurementData:
    datetime: str
    output: float

    def __post_init__(self):
        # try to convert str back to datatime to check format
        datetime.strptime(self.datetime, DATETIME_FMT)


@dataclass
class Measurement:
    data: List[MeasurementData]
    sid: str

    def to_dict(self):
        return asdict(self)



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
import datetime
from enum import Enum
from typing import List, Optional, Tuple, Union

import pandas as pd
from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.weather import Weather
from openstef_dbc.utils import (
    genereate_datetime_index,
    process_datetime_range,
    parse_influx_result,
)


class PredictorGroups(Enum):
    MARKET_DATA = "market_data"
    WEATHER_DATA = "weather_data"
    LOAD_PROFILES = "load_profiles"


class Predictor:
    def get_predictors(
        self,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: Optional[str] = None,
        location: Union[str, Tuple[float, float]] = None,
        country: str = "NL",
        source: Union[List[str], str] = "optimum",
        predictor_groups: Union[List[PredictorGroups], List[str], None] = None,
        entsoe_bidding_zone: str = "NL",
        add_weather_params: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """Get predictors.

        Get predictors for a given datetime range. Optionally predictor groups can be
        selected. If the WEATHER_DATA group is included a location is required.

        Args:
            datetime_start (datetime): Date time end.
            datetime_end (datetime): Date time start.
            location (Union[str, Tuple[float, float]], optional): Location (for weather data).
                Defaults to None.
            country (str, optional): Country code (for weather data).
                Defaults to "NL".
            source (str or list of str): which weather models should be used.
                Options: "OWM", "DSN", "WUN", "harmonie", "harm_arome", "harm_arome_fallback", "icon", "optimum",
                Default: 'optimum'. This combines harmonie, harm_arome, icon and DSN,
                taking the (heuristicly) best available source for each moment in time
            predictor_groups (Optional[List[str]], optional): The groups of predictors
                to include (see the PredictorGroups enum for allowed values). When set to
                None or not given all predictor groups will be returned. Defaults to None.
            entsoe_bidding_zone (str, optional): Name of the market place if market data is requested.
                Default to "NL".
            add_weather_params (Optional[List[str]], optional): Additional weather parameters
                / features to include. Defaults to None.

        Returns:
            pd.DataFrame: Requested predictors with timezone aware datetime index.
        """

        # Process datetimes (rounded, timezone, frequency) and generate index
        datetime_start, datetime_end, datetime_index = process_datetime_range(
            start=datetime_start,
            end=datetime_end,
            freq=forecast_resolution,
        )
        if predictor_groups is None:
            predictor_groups = [p for p in PredictorGroups]

        # convert strings to enums if required
        predictor_groups = [PredictorGroups(p) for p in predictor_groups]

        if PredictorGroups.WEATHER_DATA in predictor_groups and location is None:
            raise ValueError(
                "Need to provide a location when weather data predictors are requested."
            )
        predictors = []

        if PredictorGroups.WEATHER_DATA in predictor_groups:
            weather_data_predictors = self.get_weather_data(
                datetime_start,
                datetime_end,
                location=location,
                country=country,
                forecast_resolution=forecast_resolution,
                source=source,
                add_weather_params=add_weather_params,
            )
            predictors.append(weather_data_predictors)

        if PredictorGroups.MARKET_DATA in predictor_groups:
            market_data_predictors = self.get_market_data(
                datetime_start, datetime_end, forecast_resolution, entsoe_bidding_zone
            )
            predictors.append(market_data_predictors)

        if PredictorGroups.LOAD_PROFILES in predictor_groups:
            load_profiles_predictors = self.get_load_profiles(
                datetime_start, datetime_end, forecast_resolution
            )
            predictors.append(load_profiles_predictors)

        return pd.concat(predictors, axis=1)

    def get_market_data(
        self,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str = None,
        entsoe_bidding_zone: str = "NL",
    ) -> pd.DataFrame:
        electricity_price = self.get_day_ahead_electricity_price(
            datetime_start, datetime_end, forecast_resolution, entsoe_bidding_zone
        )

        return electricity_price

    def get_day_ahead_electricity_price(
        self,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str = None,
        entsoe_bidding_zone: str = "NL",
    ) -> pd.DataFrame:
        bind_params = {
            "_start": datetime_start,
            "_stop": datetime_end,
            "entsoe_bidding_zone": entsoe_bidding_zone,
        }

        query = f"""
            from(bucket: "forecast_latest/autogen")
                |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
                |> filter(fn: (r) => r._measurement == "marketprices" and r._field == "Price" and r.Name=="{bind_params["entsoe_bidding_zone"]}")
        """

        result = _DataInterface.get_instance().exec_influx_query(query, bind_params)

        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        # Check if response is empty
        if not result.empty:
            electricity_price = parse_influx_result(result)
        else:
            return pd.DataFrame(
                index=genereate_datetime_index(
                    start=datetime_start, end=datetime_end, freq=forecast_resolution
                )
            )
        electricity_price.rename(
            columns=dict(Price="day_ahead_electricity_price"), inplace=True
        )

        if forecast_resolution and electricity_price.empty is False:
            electricity_price = electricity_price.resample(forecast_resolution).ffill()

        return electricity_price

    def get_load_profiles(
        self,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        forecast_resolution: str = None,
    ) -> pd.DataFrame:
        """Get load profiles.

            Get the TDCV (Typical Domestic Consumption Values) load profiles from the
            database for a given range.

            MFFBAS supplies the SJI (Standaard Jaarinvoeding) and SJA (Standaard Jaarafname)
            load profiles for The Netherlands. (These load profiles are the successors of
            the SJV (Standaard Jaarverbruik) profiles from NEDU.) For more information see:
            https://www.mffbas.nl/documenten/

        Returns:
            pandas.DataFrame: TDCV load profiles (if available)

        """
        # select all fields which start with 'sjv'
        # (there is also a 'year_created' tag in this measurement)
        bind_params = {
            "_start": datetime_start,
            "_stop": datetime_end,
        }

        sjv_profles = [
            "E1A_AMI_A",
            "E1A_AMI_I",
            "E1A_AZI_A",
            "E1A_AZI_I",
            "E1B_AMI_A",
            "E1B_AMI_I",
            "E1B_AZI_A",
            "E1B_AZI_I",
            "E1C_AMI_A",
            "E1C_AMI_I",
            "E1C_AZI_A",
            "E1C_AZI_I",
            "E2A_AMI_A",
            "E2A_AMI_I",
            "E2A_AZI_A",
            "E2A_AZI_I",
            "E2B_AMI_A",
            "E2B_AMI_I",
            "E2B_AZI_A",
            "E2B_AZI_I",
            "E3A_A",
            "E3A_I",
            "E3B_A",
            "E3B_I",
            "E3C_A",
            "E3C_I",
            "E3D_A",
            "E3D_I",
            "E4A_A",
            "E4A_I",
        ]

        field_selection = '" or r["_field"] == "'.join(sjv_profles)
        query = f"""
        from(bucket: "realised/autogen")
            |> range(start: {bind_params["_start"].strftime('%Y-%m-%dT%H:%M:%SZ')}, stop: {bind_params["_stop"].strftime('%Y-%m-%dT%H:%M:%SZ')}) 
            |> filter(fn: (r) => r["_measurement"] == "sjv")
            |> filter(fn: (r) => r["_field"] == "{field_selection}")
            |> aggregateWindow(every: {forecast_resolution[:-3]}m, fn: mean, createEmpty: false)
            |> yield(name: "mean")
        
        """
        result = _DataInterface.get_instance().exec_influx_query(
            query, bind_params=bind_params
        )
        # For multiple Fields a list is returned.
        if isinstance(result, list):
            result = pd.concat(result)[["_value", "_field", "_time"]]

        if not result.empty:
            load_profiles = parse_influx_result(result)
        else:
            return pd.DataFrame(
                index=genereate_datetime_index(
                    start=datetime_start, end=datetime_end, freq=forecast_resolution
                )
            )

        if forecast_resolution and load_profiles.empty is False:
            load_profiles = load_profiles.resample(forecast_resolution).interpolate(
                limit=3
            )
        return load_profiles.shift(periods=-1, freq=forecast_resolution)

    def get_weather_data(
        self,
        datetime_start: datetime.datetime,
        datetime_end: datetime.datetime,
        location: Union[Tuple[float, float], str],
        forecast_resolution: str = None,
        country: str = "NL",
        source: Union[List[str], str] = "optimum",
        add_weather_params: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        # Get weather data
        weather_params = [
            "clouds",
            "radiation",
            "temp",
            "winddeg",
            "windspeed",
            "windspeed_100m",
            "pressure",
            "humidity",
            "rain",
            "mxlD",
            "snowDepth",
            "clearSky_ulf",
            "clearSky_dlf",
            "ssrunoff",
            *(add_weather_params or []),
        ]
        weather_data = Weather().get_weather_data(
            location,
            weather_params,
            datetime_start,
            datetime_end,
            source=source,
            country=country,
        )

        # Post process weather data
        # This might not be required anymore?
        if "source_1" in list(weather_data):
            weather_data["source"] = weather_data.source_1
            weather_data = weather_data.drop("source_1", axis=1)

        if "source" in list(weather_data):
            del weather_data["source"]

        if "input_city_1" in list(weather_data):
            del weather_data["input_city_1"]
        elif "input_city" in list(weather_data):
            del weather_data["input_city"]

        if forecast_resolution and weather_data.empty is False:
            weather_data = weather_data.resample(forecast_resolution).interpolate(
                limit=11
            )  # 11 as GFS data has data every 3 hours

        return weather_data



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Tuple, List, Dict
from datetime import datetime
import time
import re

import numpy as np
import pandas as pd

from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.log import logging
from openstef_dbc.utils import round_down_time_differences


class Write:
    def __init__(self):
        self.logger = logging.get_logger(self.__class__.__name__)

    def write_location(self, location_name: str, location: Tuple[float, float]) -> None:
        bind_params = {
            "loc": location_name,
            "lat": location[0],
            "lon": location[1],
        }

        statement = (
            "INSERT INTO NameToLatLon (regionInput, lat,lon) VALUES (:loc, :lat, :lon)"
        )

        _DataInterface.get_instance().exec_sql_write(statement, params=bind_params)

        self.logger.info("Added {location_name} to {table_name} table")

    def write_forecast(
        self,
        data: pd.DataFrame,
        dbname: str = "forecast_latest",
        influxtable: str = "prediction",
        t_ahead_series: bool = False,
        force_quality: bool = True,
    ) -> str:
        """Write a Forecast to the database.

        This function should be used to write data directly to our database.
        Do not use this function for Measurements data (pvdata, demanddata, winddata).
        Using the PostData function is preffered for this data, since traffic monitoring is ensured.

        The field 'quality' indicates if a forecasted value is:
            'actual','substituted', or 'not_renewed'

        Args:
            data (pd.DataFrame): pd.DataFrame(index="datetimeFC", columns=[
                'forecast','pid',('stdev','description','customer','quality')]
            dbname (str): The database name. Default "forecast_latest".
            influxtable (str): The table name. Default "prediction".
            t_ahead_series (bool): Should tAhead be stored to influx for specific tAheads
            force_quality (bool): If True, add a column quality with the field actual,
                or replace nan's of the column quality to actual

        Returns:
            (str): Empty if succes
        """

        # Create string placeholder for return message

        self.logger.info("Write forecast to database")

        forecast = data.copy()
        # Add created to data
        forecast["created"] = str(int(datetime.utcnow().timestamp()))

        # Add quality column. Fill nan's or missing column with 'actual'
        if force_quality:
            if "quality" not in forecast.columns:
                forecast["quality"] = "actual"
            else:
                forecast.loc[forecast["quality"].isna(), "quality"] = "actual"

        # Write DataFrame to influx database
        # Find tag columns
        tag_columns = ["pid", "type", "customer"]
        # Specify field columns
        field_columns = [x for x in forecast.columns if x not in tag_columns]

        # Cast columns to correct type, as influx is extremely picky
        casting_dict = {
            "prediction": np.float64,
            "stdev": np.float64,
            "pid": np.int64,
            "type": str,
            "customer": str,
            "created": str,
            "algtype": str,
            "description": str,
            "forecast_solar": np.float64,
            "forecast_wind_on_shore": np.float64,
            "forecast_other": np.float64,
            "forecast": np.float64,
            "quality": str,
            "tAhead": np.float64,
        }
        # Generate casting dict for available quantiles
        p = re.compile(r"quantile_P\d\d")
        quantile_columns = [s for s in field_columns if p.match(s)]
        casting_dict.update(dict.fromkeys(quantile_columns, np.float64))

        # Check if we have all columns and not some extra
        casting_dict_columns = list(casting_dict.keys())
        for k in casting_dict_columns:
            # Remove any casting dict entries that are not in the dataframe
            if k not in forecast.columns:
                casting_dict.pop(k, None)

        if set(casting_dict.keys()) != set(
            forecast.columns.to_list()
        ):  # Check if we have a type description for every column, if not raise an error
            raise ValueError(
                "Got unexpected columns, unable to cast these columns in the correct datatype."
            )

        forecast = forecast.astype(casting_dict)

        result = _DataInterface.get_instance().exec_influx_write(
            forecast.copy(),
            database=dbname,
            measurement=influxtable,
            tag_columns=tag_columns,
            field_columns=field_columns,
            time_precision="s",
        )

        message = ""

        if result:
            num_rows = str(len(forecast))
            self.logger.info(
                "Succesfully written Forecast to database",
                num_rows=num_rows,
                database=dbname,
                measurement=influxtable,
                tag_columns=tag_columns,
                field_columns=field_columns,
            )
            message += f"Written {num_rows} rows to {dbname}.{influxtable}"

        # If desired, write tAhead series to influx - tAhead table
        if t_ahead_series:
            if len(forecast) == 0:
                message += "Len forecasts=0, not going to write them to tAheads"
                return message
            message += self._write_t_ahead_series(forecast=forecast, dbname=dbname)

        return message

    def _write_t_ahead_series(
        self, forecast: pd.DataFrame, dbname: str = "forecast_latest"
    ) -> str:
        allowed_columns = [
            "tAhead",
            "pid",
            "forecast",
            "stdev",
            "customer",
            "description",
            "type",
            "quality",
        ]

        # Extract quantile column names and add to allowed columns
        quantile_forecasts = [name for name in forecast.columns if "quantile_" in name]
        allowed_columns = allowed_columns + quantile_forecasts

        influxtable = "prediction_tAheads"

        # specify desired t_aheads
        desired_t_aheads = [0.0, 1.0, 4.0, 8.0, 15.0, 24.0, 39.0, 47.0, 50.0, 144.0]

        t_adf = forecast.copy()

        # Calculate tAheads
        timediffs = (
            t_adf.index.tz_localize(None) - datetime.utcnow()
        ).total_seconds() / 3600
        # Round it to the first bigger desired_t_ahead
        t_adf["tAhead"] = round_down_time_differences(timediffs, desired_t_aheads)

        t_adf = t_adf.loc[
            [x in desired_t_aheads for x in t_adf.tAhead],
            [x for x in allowed_columns if x in t_adf.columns],
        ]
        tag_columns = ["pid", "customer", "type", "tAhead"]
        field_columns = [x for x in t_adf.columns if x not in tag_columns]

        # Force a hard typecast so floats are definetly stored as floats!
        float_columns = ["tAhead", "forecast", "stdev"] + quantile_forecasts
        float_columns_in_dataframe = [
            df_column for df_column in t_adf.columns if df_column in float_columns
        ]
        t_adf[float_columns_in_dataframe] = t_adf[float_columns_in_dataframe].apply(
            pd.to_numeric, downcast="float", errors="coerce"
        )

        result = _DataInterface.get_instance().exec_influx_write(
            t_adf.copy(),
            database=dbname,
            measurement=influxtable,
            tag_columns=tag_columns,
            field_columns=field_columns,
            time_precision="s",
        )
        if not result:
            return ""

        num_rows = str(len(t_adf))
        self.logger.info(
            "Succesfully written t ahead series to database",
            num_rows=num_rows,
            database=dbname,
            measurement=influxtable,
            tag_columns=tag_columns,
            field_columns=field_columns,
        )
        return f"Written {num_rows} rows to {dbname}.{influxtable}"

    def _write_weather_forecast_data_to_influx(
        self,
        influx_df: pd.DataFrame,
        dbname: str,
        table: str,
        tag_columns: List[str],
        field_columns: List[str],
        casting_dict: Dict[str, type],
    ):
        """Writes weather forecasts to influx database after casting the columns to the correct datatype.

        Args:
            influx_df: pd.DataFrame(index = "datetimeFC", columns = ['input_city','temp','windspeed'])
            dbname: (str) database name
            table: (str) table name
            tag_columns: (list) the column names used as tags in influx
            field_columns: (list) the column names used as fields in influx
            casting_dict: (dict) dictionary with column names as keys and the desired datatype as values

        Returns:
            None
        """
        # Check if we have all columns and not some extra
        casting_dict_columns = list(casting_dict.keys())
        for k in casting_dict_columns:
            # Remove any casting dict entries that are not in the dataframe
            if k not in influx_df.columns:
                casting_dict.pop(k, None)

        if set(casting_dict.keys()) != set(
            influx_df.columns.to_list()
        ):  # Check if we have a type description for every column, if not raise an error
            raise ValueError(
                "Got unexpected columns, unable to cast these columns in the correct datatype."
            )

        influx_df = influx_df.astype(casting_dict)

        # Write DataFrame to influx database
        result = _DataInterface.get_instance().exec_influx_write(
            influx_df.copy(),
            database=dbname,
            measurement=table,
            tag_columns=tag_columns,
            field_columns=field_columns,
            time_precision="s",
        )

        if result:
            self.logger.info(
                "Succesfully written fields to database", field_columns=field_columns
            )
            self.logger.info(
                "Successfully written tags to database", tag_columns=tag_columns
            )
            self.logger.info("Written " + str(len(influx_df)) + " rows to influx")
        else:
            self.logger.error(
                "Could not write weather data to database",
                field_columns=field_columns,
                tag_columns=tag_columns,
            )

    def _write_weather_forecast_data_latest(
        self,
        data: pd.DataFrame,
        source: str,
        forecast_created_time: datetime,
        table: str,
        dbname: str,
        tag_columns: List[str] = None,
        casting_dict: Dict[str, type] = {},
    ):
        """Writes weather data to the database. This function writes the data to a table containing the latest forecasts.

        Args:
            data: pd.DataFrame(index = "datetimeFC", columns = ['input_city','temp','windspeed'])
            source: (str) source of the weatherdata
            forecast_created_time: (datetime) the time at which the forecast was created
            table: (str) table name
            dbname: (str) database name
            tag_columns: (list) the column names used as tags in influx
            casting_dict: (dict) dictionary with column names as keys and the desired datatype as values

        Returns:
            None
        """

        influx_df = data.copy()
        influx_df["source"] = source
        # Add created to data
        influx_df["created"] = int(forecast_created_time.timestamp())

        if tag_columns is None:
            tag_columns = ["input_city", "source"]

        field_columns = [x for x in list(influx_df.columns) if x not in tag_columns]

        self._write_weather_forecast_data_to_influx(
            influx_df,
            dbname,
            table,
            tag_columns,
            field_columns,
            casting_dict,
        )

    def _write_weather_forecast_data_t_ahead(
        self,
        data: pd.DataFrame,
        source: str,
        forecast_created_time: datetime,
        table: str,
        dbname: str,
        desired_t_aheads: List[float],
        tag_columns: List[str] = None,
        casting_dict: Dict[str, type] = {},
    ):
        """Write weather data to the database. This function writes the data to a table containing with the forecasts for different t_ahead values.

        Args:
            data: pd.DataFrame(index = "datetimeFC", columns = ['input_city','temp','windspeed'])
            source: (str) source of the weatherdata
            forecast_created_time: (datetime) the time at which the forecast was created
            table: (str) table name
            dbname: (str) database name
            desired_t_aheads: (list) the t_ahead values for which the data should be written
            tag_columns: (list) the column names used as tags in influx
            casting_dict: (dict) dictionary with column names as keys and the desired datatype as values

        Returns:
            None
        """

        influx_df = data.copy()
        influx_df["source"] = source
        # Add created to data
        influx_df["created"] = int(forecast_created_time.timestamp())

        if tag_columns is None:
            tag_columns = ["input_city", "source"]
        tag_columns.append("tAhead")

        # Calculate tAheads
        timediffs = (
            influx_df.index.tz_localize(None)
            - forecast_created_time.replace(tzinfo=None)
        ).total_seconds() / 3600
        # Round it to the first bigger desired_t_ahead
        influx_df["tAhead"] = round_down_time_differences(timediffs, desired_t_aheads)

        # Remove the undesired t_aheads
        influx_df = influx_df.loc[
            [x in desired_t_aheads for x in influx_df.tAhead],
            [x for x in influx_df.columns],
        ]
        # Ignore tag columns when determining field columns
        field_columns = [x for x in influx_df.columns if x not in tag_columns]

        # Set correct casting type for tAhead column
        casting_dict["tAhead"] = np.float64

        self._write_weather_forecast_data_to_influx(
            influx_df,
            dbname,
            table,
            tag_columns,
            field_columns,
            casting_dict,
        )

    def write_weather_forecast_data(
        self,
        data: pd.DataFrame,
        source: str,
        table: str = "weather",
        dbname: str = "forecast_latest",
        tag_columns: List[str] = None,
        forecast_created_time: datetime = datetime.utcnow(),
        desired_t_aheads: List[float] = [
            1.0,
            12.0,
            15.0,
            24.0,
            36.0,
            39.0,
            48.0,
            4 * 24.0,
            6 * 24.0,
        ],
    ):
        """Write weather forecast data to the database.
        This function writes the data both to a table containing the latest forecasts,
        and a similar table with suffix _tAheads with the forecasts for different t_ahead values.

        Args:
            data: pd.DataFrame(index = "datetimeFC", columns = ['input_city','temp','windspeed'])
            source: (str) source of the weatherdata
            table: (str) table name
            dbname: (str) database name
            tag_columns: (list) the column names used as tags in influx
            forecast_created_time: (datetime) the time at which the forecast was created
            desired_t_aheads: (list) the t_ahead values for which the data should be written

        Returns:
            None
        """

        # Cast columns to correct type, as influx is extremely picky
        casting_dict = {
            "source_run": np.int64,
            "input_city": str,
            "temp": np.float64,
            "windspeed_100m": np.float64,
            "windspeed": np.float64,
            "winddeg": np.float64,
            "clouds": np.float64,
            "mxlD": np.float64,
            "snowDepth": np.float64,
            "pressure": np.float64,
            "humidity": np.float64,
            "clearSky_ulf": np.float64,
            "clearSky_dlf": np.float64,
            "radiation": np.float64,
            "radiation_direct": np.float64,
            "radiation_diffuse": np.float64,
            "radiation_normal": np.float64,
            "windspeed_100m_ensemble": np.float64,
            "windspeed_ensemble": np.float64,
            "winddeg_ensemble": np.float64,
            "clouds_ensemble": np.float64,
            "radiation_ensemble": np.float64,
            "ensemble_run": str,
            "source": str,
            "created": np.int64,
        }

        self._write_weather_forecast_data_latest(
            data=data,
            source=source,
            forecast_created_time=forecast_created_time,
            table=table,
            dbname=dbname,
            tag_columns=tag_columns,
            casting_dict=casting_dict,
        )
        self._write_weather_forecast_data_t_ahead(
            data=data,
            source=source,
            forecast_created_time=forecast_created_time,
            table=table + "_tAhead",
            dbname=dbname,
            desired_t_aheads=desired_t_aheads,
            casting_dict=casting_dict,
            tag_columns=tag_columns,
        )

    def write_realised(self, df: pd.DataFrame, sid: str):
        """Method that writes measurement data to the influx database.

        Args:
            df: pd.DataFrame(index = "datetime", columns = ['output'])
            sid: (str) String with system id of the measurement

        Returns:
            None
        """
        df["type"] = "measurement"
        df["system"] = str(sid)
        df["created"] = int(time.time())
        df = df.astype({"output": np.float64})
        # Write to influx database
        result = _DataInterface.get_instance().exec_influx_write(
            df,
            database="realised",
            measurement="power",
            tag_columns=["system", "type"],
            field_columns=["output", "created"],
            time_precision="s",
        )
        if not result:
            self.logger.error(
                "Something wend wrong while writing measurement data to influx"
            )
            return

        self.logger.info(
            "Wrote measurement data for {} systems to influx".format(
                df["system"].nunique()
            )
        )

    def write_realised_pvdata(self, df: pd.DataFrame, region: str) -> None:
        """Method that writes realised pv data to the influx database. This function
        also adds systems to the systems table in mysql if they do not yet excist.

        Args:
            df: pd.DataFrame(index = "datetime", columns = ['output','system'])
            region: (str) String with the pvdata.org region to which the systems in the df belong

        Returns:
            None
        """
        df["type"] = "solar"
        df["created"] = int(time.time())
        df = df.astype({"output": np.float64})
        # Write to influx database
        result = _DataInterface.get_instance().exec_influx_write(
            df,
            database="realised",
            measurement="power",
            tag_columns=["system", "type"],
            field_columns=["output", "created"],
            time_precision="s",
        )
        if not result:
            self.logger.error("Something wend wrong while writing pvdata to influx")
            return

        self.logger.info(
            "Wrote pv data for {} systems to influx".format(df["system"].nunique())
        )

        # Prepare dataframe for writing systems to the systems table in mysql
        systems_for_sql = df[["system"]].drop_duplicates()

        # Define placeholder string
        values = ""

        # Fill string with values
        for sid in systems_for_sql["system"]:
            values += "('" + sid + "', '" + region + "')" + ","

        # Get rid of last comma
        values = values[0:-1]

        db_type = _DataInterface.get_instance().get_sql_db_type()

        # Compose query for writing new systems
        if db_type == "mysql":
            # Compose query for writing new systems in MySQL
            query = "INSERT IGNORE INTO `systems` (sid, region) VALUES " + values
        elif db_type == "postgresql":
            # Compose query for writing new systems in PostgreSQL
            query = (
                "INSERT INTO systems (sid, region) VALUES "
                + values
                + " ON CONFLICT (sid) DO NOTHING"
            )
        else:
            self.logger.error("Unsupported database type: {}".format(db_type))
            return

        # Execute query
        _DataInterface.get_instance().exec_sql_write(query)

    def write_kpi(self, pj: dict, kpis: dict) -> None:
        """Method that writes the key performance indicators of a pid to an influx DB.
        Note that NaN / Inf are converted to 0, since these are not supported in Influx.

        Args:
            pj: a KTP prediction job
            kpis: Dictionary with dictionaries with Kpis for each t_ahead

        Returns:
            None

        """
        df = pd.DataFrame(pd.DataFrame(kpis)).T
        df["tAhead"] = [float(x.replace("h", "")) for x in df.index]
        df["pid"] = pj["id"]
        # add date
        df = df.set_index("date")

        # Convert nan / inf since these are not supported in influx
        df = df.replace([np.inf], 9999.9)
        df = df.replace([-np.inf], -9999.9)
        df = df.fillna(value=0)

        # Specify correct types
        intcols = ["pid"]
        floatcols = [x for x in df.columns if x not in intcols]
        df[floatcols] = df[floatcols].astype("float64")
        df[intcols] = df[intcols].astype("Int32")

        # Add user-friendly-fields
        df["description"] = pj["description"]
        df["customer"] = pj["name"]

        # Initialise the tag columns for Influx
        allowed_tags = ["pid", "tAhead", "description", "customer", "window_days"]
        tags = [x for x in allowed_tags if x in df.columns]

        # Write the kpis to the influx db
        _DataInterface.get_instance().exec_influx_write(
            df,
            database="forecast_latest",
            measurement="prediction_kpi",
            tag_columns=tags,
        )
        # Let user know everything went well
        self.logger.info("Succesfully wrote KPIs for pid: {}".format(str(pj["id"])))

    def write_apx_market_data(self, apx_data: pd.DataFrame) -> bool:
        success = _DataInterface.get_instance().exec_influx_write(
            apx_data,
            database="forecast_latest",
            measurement="marketprices",
            tag_columns=["Name"],
            field_columns=["Price"],
            time_precision="s",
        )
        return success

    def write_sjv_load_profiles(
        self, sjv_load_data: pd.DataFrame, field_columns: List[str]
    ) -> bool:
        success = _DataInterface.get_instance().exec_influx_write(
            sjv_load_data,
            database="realised",
            measurement="sjv",
            tag_columns=["year_created"],
            field_columns=field_columns,
            time_precision="s",
        )
        return success

    def write_windturbine_powercurves(
        self, power_curves: pd.DataFrame, if_exists: str = "fail"
    ) -> None:
        _DataInterface.get_instance().exec_sql_dataframe_write(
            power_curves, "genericpowercurves", if_exists=if_exists, index=False
        )

    def write_energy_splitting_coefficients(
        self, coefficients: dict, if_exists: str = "fail"
    ) -> None:
        _DataInterface.get_instance().exec_sql_dataframe_write(
            coefficients, "energy_split_coefs", if_exists=if_exists, index=False
        )



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""
KTP custom logging framework.

The KTP custom logging framework configures the logging functionality and enables other
modules/functions/classes to obtain a named instance. When configuring the logging
framework two arguments/options are given: the loging level and the runtime environment.
The logging level determines the lowest level of the log messages which should be displayed.
The runtime environment determines if the loggers should about a humanreadable structlog
format or a machine readable json structlog format.

In any case the logging is based on structlog (https://www.structlog.org/en/stable/).
Structlog has a different philosophy compared to more traditional logging. Struclog aims
to produce logs that are easily parsed by a machine (e.g. Kibana, Logstash etc.).

Instead of using traditional string interpolation to generate meaningfull log messages,
structlog enabled keyword arguments to given context.

traditional approach
logger.info(f"Calculate c using a = {a}, b = {b}")

structlog approach
logger.info("Calcualte c", a=a, b=b)


Example:
    # import custom logging framework
    from ktprognoses.log import logging

    # configure the logging framework
    # typically you only configure your logger once in your main/entry script
    # all other modules and classes just obtain a logger and use it
    logging.configure_logging(loglevel="INFO", runtime_env="local")

    # obtain a named logger instance
    logger = logging.get_logger("prognosis_scheduler")

    # bind a variable to the logger, from now on this variable will be added to
    # every log message, until you use `unbind`
    logger.bind(scheduler="prognosis_scheduler")

    # log something with INFO level (any number of keyword arguments may be given)
    logger.info("Start making basecase predictions", num_prediction_jobs=32)

    # prints: runtime_env == "dev"
    # 2020-08-18 20:52:35 [info     ] Start making basecase predictions [prognosis_scheduler] num_prediction_jobs=32 scheduler="prognosis_scheduler"

    # prints: runtime_env == "container"
    # {
    #   "num_prediction_jobs": 32,
    #   "event": "Start making basecase predictions",
    #   "level": "info",
    #   "logger": "prognosis_scheduler",
    #   "scheduler": "prognosis_scheduler",
    #   "timestamp": "2020-08-18 20:52:56"
    # }

    # the same principle holds for the other levels: "debug", "warning" and "error"
    # e.g. logger.debug(), logger.warning(), logger.error()

Loosely based on the work of Bas Niesink
https://github.com/Alliander/Python-API-skeleton/blob/master/app/core/initializers/logging.py
"""

import logging
import logging.config
import sys

import structlog
import warnings

import openstef_dbc.log.processors as custom_processors

__configured = False
__loggers = {}


def configure_logging(loglevel="INFO", runtime_env="local"):
    """Configure the logging functionality.

        Set-up and configure the structlog based logging functionality. The log level
        sets the lowest level which should be printed while the runtime environment
        determines if the logger outputs human readable or machine (json) output.

    Args:
        loglevel (str, int): The lowest level which should be printed
        runtime_env (str, RuntimeEnv): The runtime environment

    Returns:
        None
    """
    global __configured

    if isinstance(runtime_env, str) is False:
        try:
            runtime_env = runtime_env.value
        except AttributeError:
            raise ValueError("runtime_env need to be a string or an enum")

    if runtime_env.lower() == "local":
        _configure_logging_development(loglevel)
    elif runtime_env.lower() == "container":
        _configure_logging_deployed(loglevel)
    else:
        raise ValueError(f"Unknown runtime environment {runtime_env}")

    # Some third-party loggers can be really annoying
    # for example when using tab completions while debugging
    # raise the level of of those loggers
    if loglevel in ["DEBUG", "debug", logging.DEBUG]:
        _disable_third_party_loggers()

    __configured = True


def _disable_third_party_loggers():
    annoying_loggers_level = "ERROR"
    # parse en asyncio mess up ipython tab completion,
    # cfgrid produces a lot of irrelevant logs in de the data fetcher
    annoying_loggers = [
        "parso.python.diff",
        "parso.cache",
        "asyncio",
        "urllib3.connectionpool",
        "cfgrid.dataset",
    ]
    for name in annoying_loggers:
        logging.getLogger(name).setLevel(annoying_loggers_level)


def _configure_logging_development(loglevel):
    shared_processors = [
        # threadlocal processors
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,  # ConsoleRendered has a fixed positions
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
    ]

    processors = shared_processors + [
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter
    ]

    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    formatter = structlog.stdlib.ProcessorFormatter(
        processor=structlog.dev.ConsoleRenderer(),
        foreign_pre_chain=shared_processors,
    )

    root_logger = logging.getLogger()

    # look for an already StreamHandler
    for handler in root_logger.handlers:
        if type(handler) is logging.StreamHandler:
            handler.setFormatter(formatter)
            handler.setStream(stream=sys.stdout)
            break
    # no StreamHandler found -> add new one
    else:
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    root_logger.setLevel(loglevel)


def _configure_logging_deployed(loglevel):
    shared_processors = [
        # threadlocal processors
        structlog.contextvars.merge_contextvars,
        # stdlib processors
        structlog.stdlib.add_log_level,  # add log level to event_dict
        structlog.stdlib.add_logger_name,  # add name to event_dict
        # structlog processors
        structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"),
        # custom processors
        custom_processors.add_application_metadata,  #
        custom_processors.rename_forbidden_keys,
    ]

    pre_chain = shared_processors

    # Setup JSON-logging to stdout
    logging.captureWarnings(True)
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "json": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": structlog.processors.JSONRenderer(),
                    "foreign_pre_chain": pre_chain,
                }
            },
            "handlers": {
                "default": {
                    "level": loglevel,
                    "class": "logging.StreamHandler",
                    "formatter": "json",
                }
            },
            "loggers": {
                "": {
                    "handlers": ["default"],
                    "level": loglevel,
                    "propagate": True,
                }
            },
        }
    )

    processors = (
        [structlog.stdlib.filter_by_level]
        + shared_processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter]
    )

    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )


def get_logger(name=__name__):
    # if logging is not yet configures
    if __configured is False:
        # configure logging using defaults
        configure_logging()

    # always return the same instance for the same name
    if name not in __loggers:
        __loggers[name] = structlog.get_logger(name)

    return __loggers[name]


def initialize_logging(name, log_level=None):
    """Empty placeholder for backward compatibility"""
    warnings.warn(
        f"DEPRECATION WARNING: initialize_logging({name}) will be depreciated soon."
        f"please use the new get_logger function. log_level argument not used: {log_level}"
    )

    return get_logger(name)


def set_log_level(level):
    """Placeholder for backward compatibility"""
    if level is None or level == "None":
        level = "INFO"
    configure_logging(level)



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta
from typing import Tuple, Union, List

import numpy as np
import pandas as pd
import pytz
import structlog
from openstef_dbc.data_interface import _DataInterface
from openstef_dbc.services.ems import Ems
from openstef_dbc.services.predictor import Predictor
from openstef_dbc.services.systems import Systems
from openstef_dbc.services.weather import Weather
from openstef_dbc.utils import process_datetime_range


class ModelInput:
    def __init__(self) -> None:
        self.logger = structlog.get_logger(self.__class__.__name__)

    def get_model_input(
        self,
        pid: int = 295,
        location: Union[Tuple[int, int], str] = "Arnhem",
        country: str = "NL",
        datetime_start: str = None,
        datetime_end: str = None,
        forecast_resolution: str = "15min",
        entsoe_bidding_zone: str = "NL",
        weather_source: Union[List[str], str] = "optimum",
    ) -> pd.DataFrame:
        """Get model input.

        Get load and predictors for given pid and datetime range. If the forecast_resolution
        is lower than the data resolution, the price data is filled, while the weather
        and load data is interpolated.

        Args:
            pid (int, optional): Prediction job id. Defaults to 295.
            location (str, optional): Location name or tuple with lat, lon. Defaults to "Arnhem".
            country (str, optional): Country of were the location is located. Defaults to "NL", The Netherlands.
            datetime_start (datetime, optional): Start datetime. Defaults to None.
            datetime_end (datetime, optional): End datetime. Defaults to None.
            forecast_resolution (str, optional): Time resolution of model input
                (see pandas Date Offset frequency strings). Defaults to "15min".
            entsoe_bidding_zone (str, optional): Name of the market place if market data is requested.
                Default to "NL".
            weather_source (str or list of str, optional): which weather models should be used in preferred order.
        Returns:
            pd.DataFrame: Model input.
        """

        # TODO remove location as an argument and get location by pid from the sql database/API
        # or alternatively use a complete prediction job as input argument
        if datetime_start is None:
            datetime_start = datetime.combine(
                datetime.utcnow().date(), datetime.min.time()
            ) - timedelta(14)

        if datetime_end is None:
            datetime_end = datetime.combine(
                datetime.utcnow().date(), datetime.min.time()
            ) + timedelta(3)

        # Process datetimes (rounded, timezone, frequency) and generate index
        datetime_start, datetime_end, datetime_index = process_datetime_range(
            start=datetime_start,
            end=datetime_end,
            freq=forecast_resolution,
        )

        # Get load
        load = Ems().get_load_pid(
            pid, datetime_start, datetime_end, forecast_resolution
        )

        # Get predictors
        predictors = Predictor().get_predictors(
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            forecast_resolution=forecast_resolution,
            location=location,
            country=country,
            entsoe_bidding_zone=entsoe_bidding_zone,
            source=weather_source,
        )

        # Create model input with datetime index
        model_input = pd.DataFrame(index=datetime_index)
        model_input.index.name = "index"

        # Add load if available, else add nan column
        if not load.empty:
            load = load.resample(forecast_resolution).mean().interpolate(limit=3)
            model_input = pd.concat([model_input, load], axis=1)
        else:
            self.logger.warning("No load data returned, fill with NaN.")
            model_input["load"] = np.nan

        # Add predictors
        model_input = pd.concat([model_input, predictors], axis=1)

        return model_input

    def get_solar_input(
        self,
        location: Union[Tuple[float, float], str],
        forecast_horizon: int,
        forecast_resolution: int,
        radius: float = 0.0,
        history: int = 14,
        datetime_start: datetime = None,
        sid: str = None,
        country: str = "NL",
        source: str = "optimum",
    ) -> pd.DataFrame:
        """This function retrieves the radiation and cloud forecast for the nearest weather location
        and the relevant pvdata from a specific system or region.
        It interpolates these values according to the forecast resolution.
        Parameters:
            - engine: database connection
            - location: lat/lon values [float] or input city [str] of turbine location
            - radius: 'None' when using a specific system, range in kms when using a region
            - history: days of historic weather and pvdata used, default 14
            - forecastHorizon: length of forecast in minutes [int]
            - forecastResolution: time resolution of forecast in minutes [int]
            - datetime_start: datetime of forecast
            - source: preferred weather source as a string, default for wind is DSN
            - country (str, optional): Country of were the location is located. Defaults to "NL", The Netherlands.
            - source (str or list of str): which weather models should be used.
                Options: "OWM", "DSN", "WUN", "harmonie", "harm_arome", "harm_arome_fallback", "icon", "optimum",
                Default: 'optimum'. This combines harmonie, harm_arome, icon and DSN,
                taking the (heuristicly) best available source for each moment in time
        """
        if datetime_start is None:
            datetime_start = datetime.utcnow()

        systems_service = Systems()

        # sid selection for pvdata
        if radius == 0:
            if sid is None:
                self.logger.debug(
                    "Radius is zero and no sid was given, selecting nearest system"
                )
                systems = systems_service.get_systems_near_location(location, freq=5)
                sid = systems.loc[0].sid  # select nearest system
                self.logger.debug("Nearest system is {sid}")
        else:
            systems = systems_service.get_systems_near_location(
                location, radius, freq=5
            )
            sid = systems.sid
            self.logger.debug(f"Found {len(systems)} systems near this location")

        # Get weather data
        weather_params = ["radiation", "clouds"]
        start = datetime_start + timedelta(days=-history)  # '2017-10-10'
        end = datetime_start + timedelta(minutes=forecast_horizon)

        weather_data = Weather().get_weather_data(
            location, weather_params, start, end, source, forecast_resolution, country
        )

        # Interpolate weather data to 15 minute values
        weather_data = weather_data.resample(str(forecast_resolution) + "min").asfreq()
        for col in weather_params:
            if col in weather_data:
                weather_data.loc[:, col] = weather_data.loc[:, col].interpolate(
                    "slinear"
                )

        # Get PV_load from influx (end time at start of forecast)
        end = datetime_start
        pvdata = Ems().get_load_sid(
            sid,
            start,
            end,
            forecast_resolution,
            aggregated=True,
            average_output=radius == 0,
        )

        # If no load was found return None
        if pvdata is not None and not pvdata.empty:
            pvdata.rename(columns={"load": "aggregated"}, inplace=True)

            # Make pvdata always positive!
            pvdata["aggregated"] = pvdata["aggregated"].abs()

            # Merge data frames
            pv_data = pvdata.merge(
                weather_data, left_index=True, right_index=True, how="outer"
            )

            # Set NA values of pvdata in the past to 0
            datetime_start = pytz.utc.localize(datetime_start)
            pv_data.loc[
                (pv_data["aggregated"].isnull()) & (pv_data.index < datetime_start),
                "aggregated",
            ] = 0

            return pd.DataFrame(pv_data)

        return pd.DataFrame()

    def get_wind_input(
        self,
        location: Union[Tuple[float, float], str],
        hub_height: int,
        forecast_horizon: int,
        forecast_resolution: int,
        datetime_start: datetime = None,
        source: str = "optimum",
        country: str = "NL",
    ) -> pd.DataFrame:
        """This function retrieves the wind speed forecast for the nearest weather location
        and calculates the wind speed based on the turbine's hub height.
        It interpolates these values according to the forecast resolution.

        Args:
            location: lat/lon values [int] or input city [str] of turbine location
            forecast_horizon: length of forecast in minutes [int]
            forecast_resolution: time resolution of forecast in minutes [int]
            datetime_start: datetime of forecast
            source: preferred weather source as a string, default for wind is DSN
            country (str, optional): Country of were the location is located. Defaults to "NL", The Netherlands.
        """

        if datetime_start is None:
            datetime_start = datetime.utcnow()
        datetime_end = datetime_start + timedelta(minutes=forecast_horizon)

        windspeed = Weather().get_weather_data(
            location=location,
            weatherparams="windspeed_100m",
            datetime_start=datetime_start,
            datetime_end=datetime_end,
            source=source,
            resolution=forecast_resolution,
            country=country,
        )

        # interpolate results to 15 minute values
        windspeed = windspeed.resample(str(forecast_resolution) + "min").asfreq()
        windspeed = windspeed.interpolate("cubic")

        return pd.DataFrame(windspeed.windspeed_100m)

    @classmethod
    def get_power_curve(cls, turbine_type: str) -> dict:
        """ "This function retrieves the power curve coefficients from the genericpowercurves table,
        using the turbine type as input."""
        bind_params = {"turbine_type": turbine_type}
        query = "SELECT * FROM genericpowercurves WHERE name = :turbine_type"

        result = _DataInterface.get_instance().exec_sql_query(query, bind_params)

        if result is not None:
            result.rated_power = float(result.rated_power)
            result.slope_center = float(result.slope_center)
            result.steepness = float(result.steepness)

            return result.to_dict("records")[0]



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from functools import lru_cache
from typing import List

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class AppSettings(BaseSettings):
    """Global app settings."""

    model_config = SettingsConfigDict(
        env_prefix="openstef_", env_file=".env", extra="ignore"
    )

    optimum_weather_sources: List[str] = Field(
        [
            "harm_arome",
            "harm_arome_fallback",
            "GFS_50",
            "harmonie",
            "icon",
            "DSN",
            "ECMWF_HRES",
            "METEO_FRANCE_ARPEGE",
        ],
        description="List of weather sources to use for fetching weather data from influx using openstef DBC.",
    )


@lru_cache
def _get_app_settings() -> AppSettings:
    return AppSettings()


Settings = _get_app_settings()



# SPDX-FileCopyrightText: 2017-2022 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from dataclasses import dataclass
from datetime import datetime

DATETIME_FMT = "%Y-%m-%dT%H:%M:%S%z"


@dataclass
class SwitchState:
    datetime: str
    sid: str
    normal_open: bool
    currently_open: bool

    def __post_init__(self):
        # try to convert str back to datatime to check format
        datetime.strptime(self.datetime, DATETIME_FMT)
